#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import sys
import re
import glob
import time
import zipfile
import platform
import os
from pkg_resources import parse_version
import wx
import wx.lib.filebrowsebutton
try:
    import wx.lib.agw.hyperlink as wxhl  # 4.0+
except ImportError:
    import wx.lib.hyperlink as wxhl # <3.0.2

import psychopy
from .. import dialogs
from psychopy.localization import _translate
from psychopy import logging
from psychopy import web
import io
urllib = web.urllib

versionURL = "https://www.psychopy.org/version.txt"

"""The Updater class checks for updates and suggests that an update is carried
out if a new version is found. The actual updating is handled by
InstallUpdateDialog (via Updater.doUpdate() ).
"""


def getLatestVersionInfo(app=None):
    """
    Fetch info about the latest available version.
    Returns -1 if fails to make a connection
    """
    try:
        page = urllib.request.urlopen(versionURL)
    except urllib.error.URLError:
        return -1
    # parse update file as a dictionary
    latest = {}
    for line in page.readlines():
        # in some odd circumstances (wifi hotspots) you can fetch a
        # page that is not the correct URL but a redirect
        line = line.decode()  # convert from a byte to a str
        if line.find(':') == -1:
            return -1
            # this will succeed if every line has a key
        key, keyInfo = line.split(':')
        latest[key] = keyInfo.replace('\n', '').replace('\r', '')
    if app:
        app._latestAvailableVersion = latest
    return latest


class Updater():

    def __init__(self, app=None, runningVersion=None):
        """The updater will check for updates and download/install as needed.
        Several dialogs may be created as needed during the process.

        Usage::

            if app.prefs['AutoUpdate']:
                app.updates = Updater(app)
                app.updater.checkForUpdates()
                # if updates are found further dialogs will prompt
        """
        super(Updater, self).__init__()
        self.app = app
        if runningVersion is None:
            self.runningVersion = psychopy.__version__
        else:
            self.runningVersion = runningVersion

        # self.headers = {'User-Agent': psychopy.constants.PSYCHOPY_USERAGENT}
        self.latest = getLatestVersionInfo()
        if web.proxies is None:
            web.setupProxy()

    def getLatestInfo(self, warnMsg=False):
        # open page
        latest = getLatestVersionInfo()
        if latest == -1:
            m1 = _translate("Couldn't connect to psychopy.org to check for "
                            "updates. \n")
            m2 = _translate("Check internet settings (and proxy setting in "
                            "PsychoPy Preferences).")
            confirmDlg = dialogs.MessageDialog(
                parent=None, message=m1 + m2, type='Info',
                title=_translate('PsychoPy updates'))
            confirmDlg.ShowModal()
        return latest

    def suggestUpdate(self, confirmationDlg=False):
        """Query user about whether to update (if it's possible to update)
        """
        if self.latest is None:  # we haven't checked for updates yet
            self.latest = self.getLatestInfo()

        if self.latest == -1:
            return -1  # failed to find out about updates
        # have found 'latest'. Is it newer than running version?
        try:
            newer = self.latest['version'] > self.runningVersion
        except KeyError as err:
            print(self.latest)
            raise(err)
        skip = self.app.prefs.appData['skipVersion'] == self.latest['version']
        if newer and not skip:
            if (parse_version(self.latest['lastUpdatable'])
                    <= parse_version(self.runningVersion)):
                # go to the updating window
                confirmDlg = SuggestUpdateDialog(
                    self.latest, self.runningVersion)
                resp = confirmDlg.ShowModal()
                confirmDlg.Destroy()
                # what did the user ask us to do?
                if resp == wx.ID_CANCEL:
                    return 0  # do nothing
                if resp == wx.ID_NO:
                    self.app.prefs.appData[
                        'skipVersion'] = self.latest['version']
                    self.app.prefs.saveAppData()
                    return 0  # do nothing
                if resp == wx.ID_YES:
                    self.doUpdate()
            else:
                # the latest version needs a full install, not autoupdate
                msg = _translate("PsychoPy v%(latest)s is available (you are"
                                 " running %(running)s).\n\n")
                msg = msg % {'latest': self.latest['version'],
                             'running': self.runningVersion}
                msg += _translate("This version is too big an update to be "
                                  "handled automatically.\n")
                msg += _translate("Please fetch the latest version from "
                                  "www.psychopy.org and install manually.")
                confirmDlg = dialogs.MessageDialog(
                    parent=None, message=msg, type='Warning',
                    title=_translate('PsychoPy updates'))
                confirmDlg.cancelBtn.SetLabel(_translate('Go to downloads'))
                confirmDlg.cancelBtn.SetDefault()
                confirmDlg.noBtn.SetLabel(_translate('Go to changelog'))
                confirmDlg.yesBtn.SetLabel(_translate('Later'))
                resp = confirmDlg.ShowModal()
                confirmDlg.Destroy()
                if resp == wx.ID_CANCEL:
                    self.app.followLink(url=self.app.urls['downloads'])
                if resp == wx.ID_NO:
                    self.app.followLink(url=self.app.urls['changelog'])
        elif not confirmationDlg:  # do nothing
            return 0
        else:
            txt = _translate(
                "You are running the latest version of PsychoPy (%s). ")
            msg = txt % self.runningVersion
            confirmDlg = dialogs.MessageDialog(
                parent=None, message=msg, type='Info',
                title=_translate('PsychoPy updates'))
            confirmDlg.ShowModal()
            return -1

    def doUpdate(self):
        """Should be called from suggestUpdate
        (separate dialog to ask user whether they want to)
        """
        # app contains a reciprocal pointer to this Updater object
        dlg = InstallUpdateDialog(None, -1, app=self.app)


class SuggestUpdateDialog(wx.Dialog):
    """A dialog explaining that a new version is available
    with a link to the changelog
    """

    def __init__(self, latest, runningVersion):
        wx.Dialog.__init__(self, None, -1, title='PsychoPy3 Updates')
        sizer = wx.BoxSizer(wx.VERTICAL)

        # info about current version
        txt = _translate("PsychoPy v%(latest)s is available (you are running "
                         "%(running)s).\n\n"
                         "(To disable this check, see Preferences > "
                         "connections > checkForUpdates)")
        label = txt % {'latest': latest['version'], 'running': runningVersion}
        msg1 = wx.StaticText(self, -1, style=wx.ALIGN_CENTRE, label=label)
        if latest['lastCompatible'] > runningVersion:
            label = _translate("This version MAY require you to modify your\n"
                               "scripts/exps slightly. Read the changelog "
                               "carefully.")
            msg2 = wx.StaticText(self, -1, style=wx.ALIGN_CENTRE, label=label)
            msg2.SetForegroundColour([200, 0, 0])
        else:
            label = _translate("There are no known compatibility\nissues "
                               "with your current version.")
            msg2 = wx.StaticText(self, -1, style=wx.ALIGN_CENTRE, label=label)
        changelog = wxhl.HyperLinkCtrl(self, wx.ID_ANY,
                                       _translate("View complete Changelog"),
                                       URL="https://www.psychopy.org/changelog.html")

        if sys.platform.startswith('linux'):
            msg = _translate("You can update PsychoPy with your package "
                             "manager")
            msg3 = wx.StaticText(self, -1, msg)
        else:
            msg = _translate("Should PsychoPy update itself?")
            msg3 = wx.StaticText(self, -1, msg)

        sizer.Add(msg1, flag=wx.ALL | wx.CENTER, border=15)
        sizer.Add(msg2, flag=wx.RIGHT | wx.LEFT | wx.CENTER, border=15)
        sizer.Add(changelog, flag=wx.RIGHT | wx.LEFT | wx.CENTER, border=5)
        sizer.Add(msg3, flag=wx.ALL | wx.CENTER, border=15)

        # add buttons
        btnSizer = wx.BoxSizer(wx.HORIZONTAL)

        # for linux there should be no 'update' option
        if sys.platform.startswith('linux'):
            self.cancelBtn = wx.Button(
                self, wx.ID_CANCEL, _translate('Keep warning me'))
            self.cancelBtn.SetDefault()
            self.noBtn = wx.Button(
                self, wx.ID_NO, _translate('Stop warning me'))
        else:
            self.yesBtn = wx.Button(self, wx.ID_YES, _translate('Yes'))
            self.Bind(wx.EVT_BUTTON, self.onButton, id=wx.ID_YES)
            self.cancelBtn = wx.Button(
                self, wx.ID_CANCEL, _translate('Not now'))
            self.noBtn = wx.Button(
                self, wx.ID_NO, _translate('Skip this version'))
        self.Bind(wx.EVT_BUTTON, self.onButton, id=wx.ID_CANCEL)
        self.Bind(wx.EVT_BUTTON, self.onButton, id=wx.ID_NO)
        btnSizer.Add(self.noBtn, wx.ALIGN_LEFT)
        btnSizer.Add((60, 20), 0, wx.EXPAND)
        btnSizer.Add(self.cancelBtn)

        if not sys.platform.startswith('linux'):
            self.yesBtn.SetDefault()
            btnSizer.Add((5, 20), 0)
            btnSizer.Add(self.yesBtn)

        # configure sizers and fit
        sizer.Add(btnSizer, flag= wx.ALL, border=5)
        self.Center()
        self.SetSizerAndFit(sizer)

    def onButton(self, event):
        self.EndModal(event.GetId())


class InstallUpdateDialog(wx.Dialog):

    def __init__(self, parent, ID, app):
        """Latest is optional extra. If not given it will be fetched.
        """
        self.app = app
        # get latest version info if poss
        if app.updater in [False, None]:
            # user has turned off check for updates in prefs so check now
            app.updater = updater = Updater(app=self.app)
            # don't need a warning - we'll provide one ourselves
            self.latest = updater.getLatestInfo(warnMsg=False)
        else:
            self.latest = app.updater.latest
        self.runningVersion = app.updater.runningVersion
        wx.Dialog.__init__(self, parent, ID,
                           title=_translate('PsychoPy Updates'))

        borderSizer = wx.BoxSizer(wx.VERTICAL)
        mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(borderSizer)
        borderSizer.Add(mainSizer, border=12, proportion=1, flag=wx.ALL | wx.EXPAND)
        # set the actual content of status msg later in self.updateStatus()
        self.statusMessage = wx.StaticText(
            self, -1, "msg", style=wx.ALIGN_CENTER)
        mainSizer.Add(self.statusMessage, flag=wx.EXPAND | wx.ALL, border=5)
        # ctrls for auto-update from web
        msg = _translate(" Auto-update (will fetch latest version)")
        self.useLatestBtn = wx.RadioButton(self, -1, msg,
                                           style=wx.RB_GROUP)
        self.Bind(wx.EVT_RADIOBUTTON, self.onRadioSelect, self.useLatestBtn)
        self.progressBar = wx.Gauge(self, -1, 100, size=(250, 36))
        mainSizer.Add(self.useLatestBtn,
                      flag=wx.ALIGN_LEFT | wx.ALL, border=5)
        mainSizer.Add(self.progressBar, flag=wx.EXPAND | wx.ALL, border=5)
        # ctrls for updating from specific zip file
        msg = _translate(" Use zip file below (download a PsychoPy release "
                         "file ending .zip)")
        self.useZipBtn = wx.RadioButton(self, -1, msg)
        self.Bind(wx.EVT_RADIOBUTTON, self.onRadioSelect, self.useZipBtn)
        self.fileBrowseCtrl = wx.lib.filebrowsebutton.FileBrowseButton(
            self, -1, size=(450, 48), changeCallback=self.onFileBrowse,
            fileMask='*.zip')
        mainSizer.Add(self.useZipBtn, flag=wx.ALIGN_LEFT | wx.ALL, border=5)
        mainSizer.Add(self.fileBrowseCtrl, flag=wx.ALIGN_LEFT | wx.ALL, border=5)
        # ctrls for buttons (install/cancel)
        self.installBtn = wx.Button(self, -1, _translate('Install'))
        self.Bind(wx.EVT_BUTTON, self.onInstall, self.installBtn)
        self.installBtn.SetDefault()
        self.cancelBtn = wx.Button(self, -1, _translate('Close'))
        self.Bind(wx.EVT_BUTTON, self.onCancel, self.cancelBtn)
        btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        btnSizer.AddStretchSpacer()
        if sys.platform == "win32":
            btns = [self.installBtn, self.cancelBtn]
        else:
            btns = [self.cancelBtn, self.installBtn]
        btnSizer.Add(btns[0], 0, flag=wx.LEFT, border=5)
        btnSizer.Add(btns[1], 0, flag=wx.LEFT, border=5)
        mainSizer.AddStretchSpacer()
        mainSizer.Add(btnSizer, flag=wx.ALL | wx.EXPAND, border=5)

        self.Layout()
        self.Fit()

        # positioning and sizing
        self.updateStatus()
        self.Center()
        self.ShowModal()

    def updateStatus(self):
        """Check the current version and most recent version and update ctrls
        """
        if self.latest == -1:
            # Set message and display, and return early if version could not be found
            msg = _translate(
                "You are running PsychoPy v%s.\n ") % self.runningVersion
            msg += _translate("PsychoPy could not connect to the \n internet"
                              " to check for more recent versions.\n")
            msg += _translate("Check proxy settings in preferences.")
            self.statusMessage.SetLabel(msg)
            return
        elif (parse_version(self.latest['version'])
                  < parse_version(self.runningVersion)):
            msg = _translate(
                "You are running PsychoPy (%(running)s), which is ahead of "
                "the latest official version (%(latest)s)") % {
                'running':self.runningVersion, 'latest':self.latest['version']}
        elif self.latest['version'] == self.runningVersion:
            msg = _translate(
                "You are running the latest version of PsychoPy (%s)\n ") % self.runningVersion
            msg += _translate("You can revert to a previous version by "
                              "selecting a specific .zip source installation file")
        else:
            txt = _translate(
                "PsychoPy v%(latest)s is available\nYou are running v%(running)s")
            msg = txt % {'latest': self.latest['version'],
                         'running': self.runningVersion}
            if (parse_version(self.latest['lastUpdatable'])
                                  <= parse_version(self.runningVersion)):
                msg += _translate("\nYou can update to the latest version automatically")
            else:
                msg += _translate("\nYou cannot update to the latest version "
                                  "automatically.\nPlease fetch the latest "
                                  "Standalone package from www.psychopy.org")
        self.statusMessage.SetLabel(msg)
        areRunningLatest = self.latest['version'] == self.runningVersion
        notUpdateable = self.latest['lastUpdatable'] > self.runningVersion
        if self.latest == -1 or areRunningLatest or notUpdateable:
            # can't auto-update
            self.currentSelection = self.useZipBtn
            self.useZipBtn.SetValue(True)
            self.useLatestBtn.Disable()
        else:
            self.currentSelection = self.useLatestBtn
            self.useLatestBtn.SetValue(True)
        self.Layout()
        self.Fit()
        # this will enable/disable additional controls for the above:
        self.onRadioSelect()

    def onRadioSelect(self, event=None):
        """Set the controls of the appropriate selection to disabled/enabled
        """
        # if receive no event then just set everything to previous state
        if event != None:
            self.currentSelection = event.GetEventObject()
        else:
            pass
        if self.currentSelection == self.useLatestBtn:
            self.fileBrowseCtrl.Disable()
            self.progressBar.Enable()
        elif self.currentSelection == self.useZipBtn:
            self.fileBrowseCtrl.Enable()
            self.progressBar.Disable()
            # if this has been disabled by the fact that we couldn't connect
            self.installBtn.Enable()

    def onCancel(self, event):
        self.app.updater = None
        self.Close()

    def onFileBrowse(self, event):
        self.filename = event.GetString()

    def onInstall(self, event):
        if self.currentSelection == self.useLatestBtn:
            info = self.doAutoInstall()
        else:
            info = self.installZipFile(self.filename)
        self.statusMessage.SetLabel(str(info))
        self.Fit()

    def fetchPsychoPy(self, v='latest'):
        msg = _translate("Attempting to fetch PsychoPy %s...")
        self.statusMessage.SetLabel(msg % self.latest['version'])
        info = ""
        if v == 'latest':
            v = self.latest['version']

        # open page
        URL = "http://github.com/psychopy/psychopy/releases/download/%s/PsychoPy-%s.zip"
        page = urllib.request.urlopen(URL % v)
        # download in chunks so that we can monitor progress and abort mid-way
        chunk = 4096
        read = 0
        fileSize = int(page.info()['Content-Length'])
        buffer = io.StringIO()
        self.progressBar.SetRange(fileSize)
        while read < fileSize:
            ch = page.read(chunk)
            buffer.write(ch)
            read += chunk
            self.progressBar.SetValue(read)
            txt = _translate(
                "Fetched %(done)i of %(total)i kb of PsychoPy-%(version)s.zip")
            msg = txt % {'done': read // 1000,
                         'total': fileSize // 1000, 'version': v}
            self.statusMessage.SetLabel(msg)
            self.Update()
        info += _translate('Successfully downloaded PsychoPy-%s.zip') % v
        page.close()
        zfile = zipfile.ZipFile(buffer)
        # buffer.close()
        return zfile, info

    def installZipFile(self, zfile, v=None):
        """If v is provided this will be used as new version number;
        otherwise try and retrieve a version number from zip file name
        """
        info = ""  # return this at the end
        zfileIsName = type(zfile) == str
        if os.path.isfile(zfile) and zfileIsName:
            # zfile is filename not an actual file
            if v is None:  # try and deduce it
                zFilename = os.path.split(zfile)[-1]
                searchName = re.search(r'[0-9]*\.[0-9]*\.[0-9]*.', zFilename)
                if searchName != None:
                    v = searchName.group(0)[:-1]
                else:
                    msg = "Couldn't deduce version from zip file: %s"
                    logging.warning(msg % zFilename)
            f = open(zfile, 'rb')
            zfile = zipfile.ZipFile(f)
        else:  # assume here that zfile is a ZipFile
            pass  # todo: error checking - is it a zipfile?

        currPath = self.app.prefs.paths['psychopy']
        currVer = psychopy.__version__
        # any commands that are successfully executed may need to be undone if
        # a later one fails
        undoStr = ""
        # depending on install method, needs diff handling
        # if path ends with 'psychopy' then move it to 'psychopy-version' and
        # create a new 'psychopy' folder for new version
        # does the path contain any version number?
        versionLabelsInPath = re.findall('PsychoPy-.*/', currPath)
        # e.g. the mac standalone app, no need to refer to new version number
        onWin32 = bool(sys.platform == 'win32' and
                       int(sys.getwindowsversion()[1]) > 5)
        if len(versionLabelsInPath) == 0:
            unzipTarget = currPath
            try:  # to move existing PsychoPy
                os.rename(currPath, "%s-%s" % (currPath, currVer))
                undoStr += 'os.rename("%s-%s" %(currPath, currVer),currPath)\n'
            except Exception:
                if onWin32:
                    msg = _translate("To upgrade you need to restart the app"
                                     " as admin (Right-click the app and "
                                     "'Run as admin')")
                else:
                    msg = _translate("Could not move existing PsychoPy "
                                     "installation (permissions error?)")
                return msg
        else:  # setuptools-style installation
            # generate new target path
            unzipTarget = currPath
            for thisVersionLabel in versionLabelsInPath:
                # remove final slash from the re.findall
                pathVersion = thisVersionLabel[:-1]
                unzipTarget = unzipTarget.replace(pathVersion,
                                                  "PsychoPy-%s" % v)
                # find the .pth file that specifies the python dir
                # create the new installation directory BEFORE changing pth
                # file
                nUpdates, newInfo = self.updatePthFile(pathVersion,
                                                       "PsychoPy-%s" % v)
                if nUpdates == -1:  # there was an error (likely permissions)
                    undoStr += 'self.updatePthFile(unzipTarget, currPath)\n'
                    exec(undoStr)  # undo previous changes
                    return newInfo

        try:
            # create the new installation dir AFTER renaming existing dir
            os.makedirs(unzipTarget)
            undoStr += 'os.remove(%s)\n' % unzipTarget
        except Exception:  # revert path rename and inform user
            exec(undoStr)  # undo previous changes
            if onWin32:
                msg = _translate(
                    "Right-click the app and 'Run as admin'):\n%s")
            else:
                msg = _translate("Failed to create directory for new version"
                                 " (permissions error?):\n%s")
            return msg % unzipTarget

        # do the actual extraction
        for name in zfile.namelist():  # for each file within the zip
            # check that this file is part of psychopy (not metadata or docs)
            if name.count('/psychopy/') < 1:
                continue
            try:
                targetFile = os.path.join(unzipTarget,
                                          name.split('/psychopy/')[1])
                targetContainer = os.path.split(targetFile)[0]
                if not os.path.isdir(targetContainer):
                    os.makedirs(targetContainer)  # make the containing folder
                if targetFile.endswith('/'):
                    os.makedirs(targetFile)  # it's a folder
                else:
                    outfile = open(targetFile, 'wb')
                    outfile.write(zfile.read(name))
                    outfile.close()
            except Exception:
                exec(undoStr)  # undo previous changes
                logging.error('failed to unzip file: ' + name)
                logging.error(sys.exc_info()[0])
        info += _translate('Success. \nChanges to PsychoPy will be completed'
                           ' when the application is next run')
        self.cancelBtn.SetDefault()
        self.installBtn.Disable()
        return info

    def doAutoInstall(self, v='latest'):
        if v == 'latest':
            v = self.latest['version']
        msg = _translate("Downloading PsychoPy v%s") % v
        self.statusMessage.SetLabel(msg)
        try:
            zipFile, info = self.fetchPsychoPy(v)
        except Exception:
            msg = _translate('Failed to fetch PsychoPy release.\n'
                             'Check proxy setting in preferences')
            self.statusMessage.SetLabel(msg)
            return -1
        self.statusMessage.SetLabel(info)
        self.Fit()
        # got a download - try to install it
        info = self.installZipFile(zipFile, v)
        return info

    def updatePthFile(self, oldName, newName):
        """Searches site-packages for .pth files and replaces any instance of
        `oldName` with `newName`, expect names like PsychoPy-1.60.04
        """
        from distutils.sysconfig import get_python_lib
        siteDir = get_python_lib()
        pthFiles = glob.glob(os.path.join(siteDir, '*.pth'))
        # sometimes the site-packages dir isn't where the pth files are kept?
        enclosingSiteDir = os.path.split(siteDir)[0]
        pthFiles.extend(glob.glob(os.path.join(enclosingSiteDir, '*.pth')))
        nUpdates = 0  # no paths updated
        info = ""
        for filename in pthFiles:
            lines = open(filename, 'r').readlines()
            needSave = False
            for lineN, line in enumerate(lines):
                if oldName in line:
                    lines[lineN] = line.replace(oldName, newName)
                    needSave = True
            if needSave:
                try:
                    f = open(filename, 'w')
                    f.writelines(lines)
                    f.close()
                    nUpdates += 1
                    logging.info('Updated PsychoPy path in %s' % filename)
                except Exception:
                    info += 'Failed to update PsychoPy path in %s' % filename
                    return -1, info
        return nUpdates, info


def sendUsageStats():
    """Sends anonymous, very basic usage stats to psychopy server:
      the version of PsychoPy
      the system used (platform and version)
      the date
    """

    v = psychopy.__version__
    dateNow = time.strftime("%Y-%m-%d_%H:%M")
    miscInfo = ''

    # urllib.install_opener(opener)
    # check for proxies
    if web.proxies is None:
        web.setupProxy()

    # get platform-specific info
    if sys.platform == 'darwin':
        OSXver, junk, architecture = platform.mac_ver()
        systemInfo = "OSX_%s_%s" % (OSXver, architecture)
    elif sys.platform.startswith('linux'):
        from distro import linux_distribution
        systemInfo = '%s_%s_%s' % (
            'Linux',
            ':'.join([x for x in linux_distribution() if x != '']),
            platform.release())
        if len(systemInfo) > 30:  # if it's too long PHP/SQL fails to store!?
            systemInfo = systemInfo[0:30]
    elif sys.platform == 'win32':
        systemInfo = "win32_v" + platform.version()
    else:
        systemInfo = platform.system() + platform.release()
    u = "https://www.psychopy.org/usage.php?date=%s&sys=%s&version=%s&misc=%s"
    URL = u % (dateNow, systemInfo, v, miscInfo)
    try:
        req = urllib.request.Request(URL)
        page = urllib.request.urlopen(req)  # proxies
    except Exception:
        logging.warning("Couldn't connect to psychopy.org\n"
                        "Check internet settings (and proxy "
                        "setting in PsychoPy Preferences.")



"""Tests for the `AudioClip` class.
"""
import os
from tempfile import mkdtemp
import pytest
import numpy as np
import psychopy
from psychopy.sound import (
    AudioClip,
    AUDIO_CHANNELS_STEREO,
    AUDIO_CHANNELS_MONO,
    SAMPLE_RATE_96kHz,
    SAMPLE_RATE_48kHz,
    SAMPLE_RATE_16kHz)


@pytest.mark.audioclip
def test_audioclip_create():
    """Create an audio clip object and see if the properties are correct. Basic
    stress test to check if we get the value we expect.
    """
    nSamples = 1024
    rates = (SAMPLE_RATE_16kHz, SAMPLE_RATE_48kHz, SAMPLE_RATE_96kHz)

    for samplesRateHz in rates:
        duration = nSamples / float(samplesRateHz)
        for nChannels in (AUDIO_CHANNELS_MONO, AUDIO_CHANNELS_STEREO):
            audioClip = AudioClip(
                samples=np.zeros((nSamples, nChannels)),
                sampleRateHz=samplesRateHz)

            # check if the number of channels is correctly specified
            assert audioClip.channels == nChannels

            # check if the computed duration makes sense given the sample rate
            assert np.isclose(audioClip.duration, duration)

            # check boolean properties for channels
            if audioClip.channels == AUDIO_CHANNELS_MONO:
                assert audioClip.isMono
                # should be the same if mono already
                monoClip = audioClip.asMono()
                assert np.allclose(monoClip.samples, audioClip.samples)
            elif audioClip.channels == AUDIO_CHANNELS_STEREO:
                assert audioClip.isStereo
                # set converting to mono
                monoClip = audioClip.asMono()
                # make the clip mono inplace
                audioClip = audioClip.asMono(copy=False)
                # check if samples are the same
                assert np.allclose(monoClip.samples, audioClip.samples)


@pytest.mark.audioclip
def test_audioclip_synth():
    """Test `AudioClip` static methods for sound generation. Just check if the
    sounds created give back data structured as expected. Not testing if the
    contents are correctly generated (yet).
    """
    duration = 0.25  # quarter second long for all generated sounds
    rates = (SAMPLE_RATE_16kHz, SAMPLE_RATE_48kHz, SAMPLE_RATE_96kHz)

    # test different sample rates and channels
    for sampleRateHz in rates:
        for nChannels in (AUDIO_CHANNELS_MONO, AUDIO_CHANNELS_STEREO):
            # test white noise generation
            whiteNoise = AudioClip.whiteNoise(
                duration=duration,
                sampleRateHz=sampleRateHz,
                channels=nChannels
            )

            assert whiteNoise.channels == nChannels
            assert np.isclose(whiteNoise.duration, duration)

            # test silence
            silence = AudioClip.silence(
                duration=duration,
                sampleRateHz=sampleRateHz,
                channels=nChannels
            )

            assert silence.channels == nChannels
            assert np.isclose(silence.duration, duration)

            # test sine wave
            sineWave = AudioClip.sine(
                duration=duration,
                freqHz=440,
                gain=1.0,
                sampleRateHz=sampleRateHz,
                channels=nChannels
            )

            assert sineWave.channels == nChannels
            assert np.isclose(sineWave.duration, duration)

            # test sine wave
            squareWave = AudioClip.square(
                duration=duration,
                freqHz=440,
                dutyCycle=0.5,
                gain=1.0,
                sampleRateHz=sampleRateHz,
                channels=nChannels
            )

            assert squareWave.channels == nChannels
            assert np.isclose(squareWave.duration, duration)

            # test sine wave
            sawtoothWave = AudioClip.sawtooth(
                duration=duration,
                freqHz=440,
                peak=1.0,
                gain=1.0,
                sampleRateHz=sampleRateHz,
                channels=nChannels
            )

            assert sawtoothWave.channels == nChannels
            assert np.isclose(sawtoothWave.duration, duration)


@pytest.mark.audioclip
def test_audioclip_attrib():
    """Test `AudioClip` attribute setters and getters. Tests attributes
    `samples`, `sampleRateHz`, `duration`, and `gain()`.
    """
    # generate an audio clip using a sine wave
    originalDuration = 1.0
    originalSampleRateHz = SAMPLE_RATE_48kHz
    audioClip = AudioClip.sine(
        duration=originalDuration,
        sampleRateHz=originalSampleRateHz,
        gain=0.8
    )

    # check if changing the sample rate results in a change in duration
    audioClip.sampleRateHz = SAMPLE_RATE_16kHz

    # should compute as longer as the same number of samples being sampled
    # slower should result in a longer duration
    assert audioClip.duration > originalDuration

    # check if the change in duration has the correct ratio given the new rate
    assert np.isclose(audioClip.duration, SAMPLE_RATE_48kHz / SAMPLE_RATE_16kHz)

    # check if setting samples works, just halve the number of samples and check
    # if the new duration is half as long
    audioClip.sampleRateHz = SAMPLE_RATE_48kHz  # reset
    trimAt = int(audioClip.samples.shape[0] / 2.0)
    audioClip.samples = audioClip.samples[:trimAt, :]
    assert np.isclose(originalDuration / 2.0, audioClip.duration)

    # test if gain works, for our original data, not value should be above 0.8
    assert np.max(audioClip.samples) <= 0.81 and \
           np.min(audioClip.samples) >= -0.81

    # apply gain to max and retest
    audioClip.gain(0.2)  # 20% increase in gain
    assert np.max(audioClip.samples) <= 1.0 and \
           np.min(audioClip.samples) >= -1.0

    # give bad value to gain for the channel
    caughtChannelValueError = False
    try:
        audioClip.gain(1.0, channel=2)
    except ValueError:
        caughtChannelValueError = True

    assert caughtChannelValueError, \
        "Failed to catch error be specifying wrong number to `channel` param " \
        "in `.gain()`."


@pytest.mark.audioclip
def test_audioclip_concat():
    """Test combining audio clips together.
    """
    # durations to use for each segment
    dur1, dur2, dur3 = 0.2, 0.1, 0.7
    totalDur = sum([dur1, dur2, dur3])

    # create a bunch of clips to join
    clip1 = AudioClip.silence(duration=dur2, sampleRateHz=SAMPLE_RATE_48kHz)
    clip2 = AudioClip.whiteNoise(duration=dur1, sampleRateHz=SAMPLE_RATE_48kHz)
    clip3 = AudioClip.sine(duration=dur3, sampleRateHz=SAMPLE_RATE_48kHz)

    # concatenate clips using the `+` operator
    newClip1 = clip1 + clip2 + clip3

    # check the new duration
    assert np.isclose(newClip1.duration, totalDur)
    assert np.isclose(newClip1.samples.shape[0], SAMPLE_RATE_48kHz)

    # do the same using the append() method
    newClip2 = (clip1.copy()).append(clip2).append(clip3)

    assert np.isclose(newClip2.duration, totalDur)
    assert np.isclose(newClip2.samples.shape[0], SAMPLE_RATE_48kHz)

    # assert that these two methods do the same thing
    assert np.allclose(newClip1.samples, newClip2.samples)

    # test copy
    newClip3 = clip1.copy()
    assert np.allclose(newClip3.samples, clip1.samples)

    # test inplace
    originalObjectId = id(newClip3)
    newClip3 += clip2
    newClip3 += clip3

    # make sure the object is the same in this case
    assert id(newClip3) == originalObjectId
    # check if samples are the same as above
    assert np.allclose(newClip3.samples, newClip2.samples)
    assert np.isclose(newClip3.samples.shape[0], SAMPLE_RATE_48kHz)

    # ensure that concatenation fails when sample rates are heterogeneous
    clipBad = AudioClip.whiteNoise(
        duration=dur1,
        sampleRateHz=SAMPLE_RATE_16kHz)

    caughtSampleRateError = False
    try:
        _ = clipBad + clip1
    except AssertionError:
        caughtSampleRateError = True

    assert caughtSampleRateError, \
        "Did not catch expected error when combining `AudioClip` objects " \
        "with heterogeneous sample rates."

    # check what happens when clips are empty when combined
    emptyClip = AudioClip(np.zeros((0, 0)), sampleRateHz=SAMPLE_RATE_48kHz)

    # check if the empty clip contains the data from the other
    clipData = clip1.copy()
    newClip4 = emptyClip.append(clipData)
    assert np.allclose(clipData.samples, newClip4.samples)

    # other direction
    newClip4 = (clipData.copy()).append(emptyClip)
    assert np.allclose(clipData.samples, newClip4.samples)


@pytest.mark.audioclip
def test_audioclip_file():
    """Test saving and loading audio samples from files. Checks the integrity
    of loaded data to ensure things are similar to the original.
    """
    # generate samples
    np.random.seed(123456)

    rates = (SAMPLE_RATE_16kHz, SAMPLE_RATE_48kHz, SAMPLE_RATE_96kHz)
    for sampleRateHz in rates:
        for nChannels in (AUDIO_CHANNELS_MONO, AUDIO_CHANNELS_STEREO):
            audioClip = AudioClip.whiteNoise(
                duration=1.0, sampleRateHz=sampleRateHz, channels=nChannels)

            # create temporary folder for data
            testDir = mkdtemp(prefix='psychopy-tests-test_audioclip')

            # save as WAV file
            fname = os.path.join(testDir, 'test_audioclip_file.wav')
            audioClip.save(fname)

            # load it
            loadedAudioClip = AudioClip.load(fname)

            # check if they are the same, there is some error from quantization
            assert np.allclose(
                loadedAudioClip.samples,
                audioClip.samples,
                atol=1e-4)
            assert loadedAudioClip.channels == nChannels

            # save and load again
            loadedAudioClip.save(fname)
            loadedAudioClip2 = AudioClip.load(fname)

            # quantization applied, lossy but should be stable here
            assert np.allclose(
                loadedAudioClip.samples,
                loadedAudioClip2.samples)
            assert loadedAudioClip2.channels == nChannels


@pytest.mark.audioclip
def test_audioclip_rms():
    """Test the RMS method of `AudioClip`. Just check if the function give back
    values that are correctly formatted given the input data.
    """
    # test clip
    audioClipStereo = AudioClip.sine(
        duration=1.0,
        sampleRateHz=SAMPLE_RATE_48kHz,
        channels=AUDIO_CHANNELS_STEREO)

    # check error when channels are specified incorrectly
    caughtChannelParamError = False
    try:
        audioClipStereo.rms(-1)
    except AssertionError:
        caughtChannelParamError = True

    assert caughtChannelParamError, \
        "Did not catch expected error related to specifying the wrong value " \
        "when specifying `channel` to RMS."

    # should get back 2 values
    rmsResultStereo = audioClipStereo.rms()
    assert isinstance(rmsResultStereo, np.ndarray) and \
           len(rmsResultStereo) == audioClipStereo.channels

    # make it mono, do it again
    audioClipMono = audioClipStereo.asMono()
    rmsResultMono = audioClipMono.rms()  # should be float for one channel
    assert isinstance(rmsResultMono, np.float32)


if __name__ == "__main__":
    # runs if this script is directly executed
    test_audioclip_create()
    test_audioclip_synth()
    test_audioclip_attrib()
    test_audioclip_concat()
    test_audioclip_file()
    test_audioclip_rms()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import tempfile

from . import PavloviaMiniBrowser
from .functions import logInPavlovia, logOutPavlovia
from psychopy.localization import _translate
from psychopy.projects import pavlovia
from psychopy.app import utils
import requests
import io
from psychopy import prefs
import os
import wx
import wx.lib.statbmp

from ...projects.pavlovia import User

try:
    import wx.lib.agw.hyperlink as wxhl  # 4.0+
except ImportError:
    import wx.lib.hyperlink as wxhl # <3.0.2


class UserFrame(wx.Dialog):
    def __init__(self, parent,
                 style=wx.DEFAULT_DIALOG_STYLE | wx.CENTER | wx.TAB_TRAVERSAL | wx.RESIZE_BORDER,
                 size=(-1, -1)):

        wx.Dialog.__init__(self, parent,
                           style=style,
                           size=size)
        self.app = parent.app
        self.sizer = wx.BoxSizer()
        self.sizer.Add(UserPanel(self), proportion=1, border=12, flag=wx.ALL | wx.EXPAND)
        self.SetSizerAndFit(self.sizer)


class UserPanel(wx.Panel):
    def __init__(self, parent,
                 size=(600, 300),
                 style=wx.NO_BORDER):
        wx.Panel.__init__(self, parent, -1,
                          size=size,
                          style=style)
        self.parent = parent
        self.SetBackgroundColour("white")
        self.session = pavlovia.getCurrentSession()
        iconCache = parent.app.iconCache
        # Setup sizer
        self.contentBox = wx.BoxSizer()
        self.SetSizer(self.contentBox)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.contentBox.Add(self.sizer, proportion=1, border=12, flag=wx.ALL | wx.EXPAND)
        # Head sizer
        self.headSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.headSizer, border=0, flag=wx.EXPAND)
        # Icon
        self.icon = wx.lib.statbmp.GenStaticBitmap(self, ID=wx.ID_ANY, bitmap=iconCache.getBitmap(name="user_none", size=128), size=(128, 128))
        self.icon.SetBackgroundColour("#f2f2f2")
        self.headSizer.Add(self.icon, border=6, flag=wx.ALL)
        # Title sizer
        self.titleSizer = wx.BoxSizer(wx.VERTICAL)
        self.headSizer.Add(self.titleSizer, proportion=1, flag=wx.EXPAND)
        # Full name
        self.fullName = wx.StaticText(self, size=(-1, -1), label="---", style=wx.ST_ELLIPSIZE_END)
        self.fullName.SetFont(
            wx.Font(24, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD)
        )
        self.fullName.Bind(wx.EVT_TEXT, self.updateUser)
        self.titleSizer.Add(self.fullName, border=6, flag=wx.ALL | wx.EXPAND)
        # Organisation
        self.organisation = wx.StaticText(self, size=(-1, -1), label="---", style=wx.ST_ELLIPSIZE_END)
        self.organisation.SetFont(
            wx.Font(12, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_ITALIC, wx.FONTWEIGHT_NORMAL)
        )
        self.titleSizer.Add(self.organisation, border=6, flag=wx.ALL | wx.EXPAND)
        # Spacer
        self.titleSizer.AddStretchSpacer(1)
        # Button sizer
        self.btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.titleSizer.Add(self.btnSizer, border=6, flag=wx.TOP | wx.BOTTOM | wx.EXPAND)
        # Spacer
        self.btnSizer.AddStretchSpacer(1)
        # Pavlovia link
        self.link = wxhl.HyperLinkCtrl(self, -1,
                                       label="",
                                       URL="https://gitlab.pavlovia.org/",
                                       )
        self.link.SetBackgroundColour(self.GetBackgroundColour())
        self.btnSizer.Add(self.link, border=6, flag=wx.RIGHT | wx.ALIGN_CENTER_VERTICAL)
        # Edit
        self.edit = wx.Button(self, label=chr(int("270E", 16)), size=(24, -1))
        self.edit.Bind(wx.EVT_BUTTON, self.onEdit)
        self.btnSizer.Add(self.edit, border=3, flag=wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL)
        # Login
        self.login = wx.Button(self, label=_translate("Login"))
        self.login.SetBitmap(iconCache.getBitmap(name="person_off", size=16))
        self.login.Bind(wx.EVT_BUTTON, self.onLogin)
        self.btnSizer.Add(self.login, border=3, flag=wx.LEFT | wx.EXPAND)
        # Logout
        self.logout = wx.Button(self, label=_translate("Logout"))
        self.logout.SetBitmap(iconCache.getBitmap(name="person_off", size=16))
        self.logout.Bind(wx.EVT_BUTTON, self.onLogout)
        self.btnSizer.Add(self.logout, border=3, flag=wx.LEFT | wx.EXPAND)
        # Sep
        self.sizer.Add(wx.StaticLine(self, -1), border=6, flag=wx.EXPAND | wx.ALL)
        # Bio
        self.description = wx.StaticText(self, size=(-1, -1), label="", style=wx.TE_MULTILINE)
        self.sizer.Add(self.description, proportion=1, border=6, flag=wx.ALL | wx.EXPAND)

        # Populate
        self.user = self.user

    @property
    def user(self):
        if not hasattr(self, "_user"):
            self._user = self.session.user
        return self._user

    @user.setter
    def user(self, user):
        self._user = user

        if user is None:
            iconCache = self.parent.app.iconCache
            # Icon
            self.icon.SetBitmap(iconCache.getBitmap(name="user_none", size=128))
            self.icon.Disable()
            # Full name
            self.fullName.SetLabelText("---")
            self.fullName.Disable()
            # Organisation
            self.organisation.SetLabelText("---")
            self.organisation.Disable()
            # Link
            self.link.SetLabel("")
            self.link.SetURL("https://pavlovia.org/")
            self.link.Disable()
            # Hide logout and show login
            self.logout.Hide()
            self.login.Show()
            # Disable edit
            self.edit.Disable()
            # Description
            self.description.SetLabelText("")
            self.description.Disable()
        else:
            try:
                content = requests.get(user['avatar_url']).content
                buffer = wx.Image(io.BytesIO(content))
                buffer = buffer.Scale(*self.icon.Size, quality=wx.IMAGE_QUALITY_HIGH)
                icon = wx.Bitmap(buffer)
            except requests.exceptions.MissingSchema:
                icon = wx.Bitmap()
            self.icon.SetBitmap(icon)
            self.icon.Enable()
            # Full name
            self.fullName.SetLabelText(user.user.attributes['name'] or "")
            self.fullName.Enable()
            # Organisation
            self.organisation.SetLabelText(user['organization'] or "No organization")
            self.organisation.Enable()
            # Link
            self.link.SetLabel(user['username'])
            self.link.SetURL(user['web_url'] or "")
            self.link.Enable()
            # Hide logout and show login
            self.logout.Show()
            self.login.Hide()
            # Enable edit
            self.edit.Enable()
            # Description
            self.description.SetLabelText(user['bio'] or "")
            self.description.Wrap(self.Size[0] - 36)
            self.description.Enable()
        self.Layout()

    def onLogout(self, evt=None):
        self.user = logOutPavlovia(self.parent)

    def onLogin(self, evt=None):
        self.user = logInPavlovia(parent=self.parent)

    def onEdit(self, evt=None):
        # Open edit window
        dlg = PavloviaMiniBrowser(parent=self, loginOnly=False)
        dlg.editUserPage()
        dlg.ShowModal()
        # Refresh user on close
        self.user = User(self.user.id)

    def updateUser(self, evt=None):
        """
        Disabled for now, as fields are not editable.
        """
        # Skip if no user
        if self.user is None or evt is None or self.session.user is None:
            return
        # Get object
        obj = evt.GetEventObject()
        # Update full name
        if obj == self.fullName:
            self.user.name = self.fullName.GetValue()
            #self.user.save()
        # Update organisation
        if obj == self.organisation:
            self.user.organization = self.organisation.GetValue()
            #self.user.save()
        # Update bio
        if obj == self.description:
            self.user.bio = self.description.GetValue()
            #self.user.save()
        # Update avatar
        if obj == self.icon:
            # Create temporary image file
            _, temp = tempfile.mkstemp(suffix=".png")
            self.icon.BitmapFull.SaveFile(temp, wx.BITMAP_TYPE_PNG)
            # Load and upload from temp file
            self.user.avatar = open(temp, "rb")
            #self.user.save()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of color spaces: RGB, DKL, HSV

Note that for this demo to present colors properly in calibrated
DKL space (where isoluminant stimuli have elevation=0) you need
to calibrate your monitor with a suitable spectrophotometer. If you
have a PR60 then you can do this automatically using MonitorCenter.py
in the monitors package

Note for each stimulus that the color refers to the central bar on the grating
If there are multiple lobes (a high enough SF) then the other color is simply the
complement of the one specified (passing through neutral gray)
"""
from psychopy import visual, event, core

win = visual.Window(size=(600, 600), units='norm')

stims = []
# rgb colors
stims.append( visual.GratingStim(win, mask='gauss', color='red', pos=[-0.5, 0.5], sf=2) )  # r
stims.append( visual.GratingStim(win, mask='gauss', color=(-1, 1, -1), colorSpace='rgb', pos=[-0.5, 0], sf=2))  # g
stims.append( visual.GratingStim(win, mask='gauss', color=(0, 0, 255), colorSpace='rgb255', pos=[-0.5, -0.5], sf=2))  # b

# DKL cardinal axes (see Derrington, Krauskopf and Lennie 1986)
stims.append( visual.GratingStim(win, mask='gauss', color=(90, 0, 1), colorSpace='dkl', pos=[0, 0.5], sf=2) )  # achrom
stims.append( visual.GratingStim(win, mask='gauss', color=(0, 0, 1), colorSpace='dkl', pos=[0, 0], sf=2))  # L-M
stims.append( visual.GratingStim(win, mask='gauss', color=(0, 90, 1), colorSpace='dkl', pos=[0, -0.5], sf=2))  # S

# cone-isolating stimuli
# stims.append( visual.GratingStim(win, mask='gauss', color=(0.2, 0, 0), colorSpace='lms', pos=[0.5, 0.5], sf=2))
# stims.append( visual.GratingStim(win, mask='gauss', color=(0, 0.2, 0), colorSpace='lms', pos=[0.5, 0], sf=2))
# stims.append( visual.GratingStim(win, mask='gauss', color=(0, 0, 0.5), colorSpace='lms', pos=[0.5, -0.5], sf=2))

# HSV. This is a device-dependent space
# (i.e. it will differ on each monitor but needs no calibration)
stims.append( visual.GratingStim(win, mask='gauss', color=(0, 1, 1), colorSpace='hsv', pos=[0.5, 0.5], sf=2))
stims.append( visual.GratingStim(win, mask='gauss', color=(45, 1, 1), colorSpace='hsv', pos=[0.5, 0], sf=2))
stims.append( visual.GratingStim(win, mask='gauss', color=(90, 1, 1), colorSpace='hsv', pos=[0.5, -0.5], sf=2))

labels = visual.TextStim(win, text='RGB          DKL          HSV', pos=(0,.85), wrapWidth=2)

for thisStim in stims:
    thisStim.draw()
labels.draw()
win.flip()

clock = core.Clock()
while not event.getKeys() and clock.getTime() < 10:
    core.wait(.2)

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""To handle input from keyboard (supersedes event.getKeys)


The Keyboard class was new in PsychoPy 3.1 and replaces the older
`event.getKeys()` calls.

Psychtoolbox versus event.getKeys
------------------------------------

On 64 bits Python3 installations it provides access to the
`Psychtoolbox kbQueue <http://psychtoolbox.org/docs/KbQueueCreate>`_ series of
functions using the same compiled C code (available in python-psychtoolbox lib).

On 32 bit installations and Python2 it reverts to the older
:func:`psychopy.event.getKeys` calls.

The new calls have several advantages:

- the polling is performed and timestamped asynchronously with the main thread
  so that times relate to when the key was pressed, not when the call was made
- the polling is direct to the USB HID library in C, which is faster than
  waiting for the operating system to poll and interpret those same packets
- we also detect the KeyUp events and therefore provide the option of returning
  keypress duration
- on Linux and Mac you can also distinguish between different keyboard devices
  (see :func:`getKeyboards`)

This library makes use, where possible of the same low-level asynchronous
hardware polling as in `Psychtoolbox <http://psychtoolbox.org/>`_

.. currentmodule:: psychopy.hardware.keyboard

Example usage

------------------------------------

.. code-block:: python

    from psychopy.hardware import keyboard
    from psychopy import core

    kb = keyboard.Keyboard()

    # during your trial
    kb.clock.reset()  # when you want to start the timer from
    keys = kb.getKeys(['right', 'left', 'quit'], waitRelease=True)
    if 'quit' in keys:
        core.quit()
    for key in keys:
        print(key.name, key.rt, key.duration)

"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from collections import deque
import sys
import copy
import psychopy.core
import psychopy.clock
from psychopy import logging
from psychopy.constants import NOT_STARTED
import time

try:
    import psychtoolbox as ptb
    from psychtoolbox import hid
    havePTB = True

except ImportError as err:
    logging.warning(("Import Error: "
                     + err.args[0]
                     + ". Using event module for keyboard component."))
    from psychopy import event
    havePTB = False

defaultBufferSize = 10000
# default ptb flush_type, used by macOS & linux
_ptb_flush_type = 1

# monkey-patch bug in PTB keyboard where winHandle=0 is documented but crashes.
# Also set ptb _ptb_flush_type to 0 for win32.
if havePTB and sys.platform == 'win32':
    from psychtoolbox import PsychHID
    # make a new function where we set default win_handle to be None instead of 0
    def _replacement_create_queue(self, num_slots=10000, flags=0, win_handle=None):
        PsychHID('KbQueueCreate', self.device_number,
                 None, 0, num_slots, flags, win_handle)
    # replace the broken function with ours
    hid.Keyboard._create_queue = _replacement_create_queue

    # On win32, flush_type must be 0 or events can get flushed before being processed
    _ptb_flush_type = 0


def getKeyboards():
    """Get info about the available keyboards.

    Only really useful on Mac/Linux because on these the info can be used to
    select a particular physical device when calling :class:`Keyboard`. On Win
    this function does return information correctly but the :class:Keyboard
    can't make use of it.

    Returns
    ----------
    A list of dicts
        USB Info including with name, manufacturer, id, etc for each device

    """
    if havePTB:
        indices, names, keyboards = hid.get_keyboard_indices()
        return keyboards
    return []


class Keyboard:
    """The Keyboard class provides access to the Psychtoolbox KbQueue-based
    calls on **Python3 64-bit** with fall-back to `event.getKeys` on legacy
    systems.

    """
    _backend = None
    _iohubKeyboard = None
    _iohubOffset = 0.0

    def __init__(self, device=-1, bufferSize=10000, waitForStart=False, clock=None, backend=None):
        """Create the device (default keyboard or select one)

        Parameters
        ----------
        device: int or dict

            On Linux/Mac this can be a device index
            or a dict containing the device info (as from :func:`getKeyboards`)
            or -1 for all devices acting as a unified Keyboard

        bufferSize: int

            How many keys to store in the buffer (before dropping older ones)

        waitForStart: bool (default False)

            Normally we'll start polling the Keyboard at all times but you
            could choose not to do that and start/stop manually instead by
            setting this to True

        """
        global havePTB

        if self._backend is None and backend in ['iohub', 'ptb', 'event', '']:
            Keyboard._backend = backend

        if self._backend is None:
            Keyboard._backend = ''

        if backend and self._backend != backend:
            logging.warning("keyboard.Keyboard already using '%s' backend. Can not switch to '%s'" % (self._backend,
                                                                                                      backend))

        self.status = NOT_STARTED
        # Initiate containers for storing responses
        self.keys = []  # the key(s) pressed
        self.corr = 0  # was the resp correct this trial? (0=no, 1=yes)
        self.rt = []  # response time(s)
        self.time = []  # Epoch

        if clock:
            self.clock = clock
        else:
            self.clock = psychopy.clock.Clock()

        if Keyboard._backend in ['', 'iohub']:
            from psychopy.iohub.client import ioHubConnection
            from psychopy.iohub.devices import Computer
            if not ioHubConnection.getActiveConnection() and Keyboard._backend == 'iohub':
                # iohub backend was explicitly requested, but iohub is not running, so start it up
                # setting keyboard to use standard psychopy key mappings
                from psychopy.iohub import launchHubServer
                launchHubServer(Keyboard=dict(use_keymap='psychopy'))

            if ioHubConnection.getActiveConnection() and Keyboard._iohubKeyboard is None:
                Keyboard._iohubKeyboard = ioHubConnection.getActiveConnection().getDevice('keyboard')
                Keyboard._iohubOffset = Computer.global_clock.getLastResetTime()
                Keyboard._backend = 'iohub'

        if Keyboard._backend in ['', 'ptb'] and havePTB:
            Keyboard._backend = 'ptb'
            # get the necessary keyboard buffer(s)
            if sys.platform == 'win32':
                self._ids = [-1]  # no indexing possible so get the combo keyboard
            else:
                allInds, allNames, allKBs = hid.get_keyboard_indices()
                if device == -1:
                    self._ids = allInds
                elif type(device) in [list, tuple]:
                    self._ids = device
                else:
                    self._ids = [device]

            self._buffers = {}
            self._devs = {}
            for devId in self._ids:
                # now we have a list of device IDs to monitor
                if devId == -1 or devId in allInds:
                    buffer = _keyBuffers.getBuffer(devId, bufferSize)
                    self._buffers[devId] = buffer
                    self._devs[devId] = buffer.dev

            # Is this right, waiting if waitForStart=False??
            if not waitForStart:
                self.start()

        if Keyboard._backend in ['', 'event']:
            global event
            from psychopy import event
            Keyboard._backend = 'event'

        logging.info('keyboard.Keyboard is using %s backend.' % Keyboard._backend)

    @classmethod
    def getBackend(self):
        """Return backend being used."""
        return self._backend

    @classmethod
    def setBackend(self, backend):
        """
        Set backend event handler. Returns currently active handler.

        :param backend: 'iohub', 'ptb', 'event', or ''
        :return: str
        """
        if self._backend is None:
            if backend in ['iohub', 'ptb', 'event', '']:
                Keyboard._backend = backend
            else:
                logging.warning("keyboard.Keyboard.setBackend failed. backend must be one of %s"
                                % str(['iohub', 'ptb', 'event', '']))
            if backend == 'event':
                global event
                from psychopy import event
        else:
            logging.warning("keyboard.Keyboard.setBackend already using '%s' backend. "
                            "Can not switch to '%s'" % (self._backend, backend))

        return self._backend

    def start(self):
        """Start recording from this keyboard """
        if Keyboard._backend == 'ptb':
            for buffer in self._buffers.values():
                buffer.start()

    def stop(self):
        """Start recording from this keyboard"""
        if Keyboard._backend == 'ptb':
            logging.warning("Stopping key buffers but this could be dangerous if"
                            "other keyboards rely on the same.")
            for buffer in self._buffers.values():
                buffer.stop()

    def getKeys(self, keyList=None, waitRelease=True, clear=True):
        """

        Parameters
        ----------
        keyList: list (or other iterable)

            The keys that you want to listen out for. e.g. ['left', 'right', 'q']

        waitRelease: bool (default True)

            If True then we won't report any "incomplete" keypress but all
            presses will then be given a `duration`. If False then all
            keys will be presses will be returned, but only those with a
            corresponding release will contain a `duration` value (others will
            have `duration=None`

        clear: bool (default True)

            If False then keep the keypresses for further calls (leave the
            buffer untouched)

        Returns
        -------
        A list of :class:`Keypress` objects

        """
        keys = []
        if Keyboard._backend == 'ptb':
            for buffer in self._buffers.values():
                for origKey in buffer.getKeys(keyList, waitRelease, clear):
                    # calculate rt from time and self.timer
                    thisKey = copy.copy(origKey)  # don't alter the original
                    thisKey.rt = thisKey.tDown - self.clock.getLastResetTime()
                    keys.append(thisKey)
        elif Keyboard._backend == 'iohub':
            watchForKeys = keyList
            if waitRelease:
                key_events = Keyboard._iohubKeyboard.getReleases(keys=watchForKeys, clear=clear)
            else:
                key_events = Keyboard._iohubKeyboard.getPresses(keys=watchForKeys, clear=clear)

            for k in key_events:
                kname = k.key

                if waitRelease:
                    tDown = k.time-k.duration
                else:
                    tDown = k.time

                kpress = KeyPress(code=None, tDown=tDown, name=kname)
                kpress.rt = kpress.tDown - self.clock.getLastResetTime() + Keyboard._iohubOffset
                if waitRelease:
                    kpress.duration = k.duration

                keys.append(kpress)
        else:  # Keyboard.backend == 'event'
            global event
            name = event.getKeys(keyList, modifiers=False, timeStamped=False)
            rt = self.clock.getTime()
            if len(name):
                thisKey = KeyPress(code=None, tDown=rt, name=name[0])
                keys.append(thisKey)
        return keys

    def waitKeys(self, maxWait=float('inf'), keyList=None, waitRelease=True,
                 clear=True):
        """Same as `~psychopy.hardware.keyboard.Keyboard.getKeys`, 
        but halts everything (including drawing) while awaiting keyboard input.
    
        :Parameters:
            maxWait : any numeric value.
                Maximum number of seconds period and which keys to wait for.
                Default is float('inf') which simply waits forever.
            keyList : **None** or []
                Allows the user to specify a set of keys to check for.
                Only keypresses from this set of keys will be removed from
                the keyboard buffer. If the keyList is `None`, all keys will be
                checked and the key buffer will be cleared completely.
                NB, pygame doesn't return timestamps (they are always 0)
            waitRelease: **True** or False
                If True then we won't report any "incomplete" keypress but all
                presses will then be given a `duration`. If False then all
                keys will be presses will be returned, but only those with a
                corresponding release will contain a `duration` value (others will
                have `duration=None`
            clear : **True** or False
                Whether to clear the keyboard event buffer (and discard preceding
                keypresses) before starting to monitor for new keypresses.
    
        Returns None if times out.
    
        """
        timer = psychopy.core.Clock()

        if clear:
            self.clearEvents()

        while timer.getTime() < maxWait:
            keys = self.getKeys(keyList=keyList, waitRelease=waitRelease, clear=clear)
            if keys:
                return keys
            time.sleep(0.00001)

        logging.data('No keypress (maxWait exceeded)')
        return None

    def clearEvents(self, eventType=None):
        """"""
        if Keyboard._backend == 'ptb':
            for buffer in self._buffers.values():
                buffer.flush()  # flush the device events to the soft buffer
                buffer._evts.clear()
                buffer._keys.clear()
                buffer._keysStillDown.clear()
        elif Keyboard._backend == 'iohub':
            Keyboard._iohubKeyboard.clearEvents()
        else:
            global event
            event.clearEvents(eventType)


class KeyPress(object):
    """Class to store key presses, as returned by `Keyboard.getKeys()`

    Unlike keypresses from the old event.getKeys() which returned a list of
    strings (the names of the keys) we now return several attributes for each
    key:

        .name: the name as a string (matching the previous pyglet name)
        .rt: the reaction time (relative to last clock reset)
        .tDown: the time the key went down in absolute time
        .duration: the duration of the keypress (or None if not released)

    Although the keypresses are a class they will test `==`, `!=` and `in`
    based on their name. So you can still do::

        kb = KeyBoard()
        # wait for keypresses here
        keys = kb.getKeys()
        for thisKey in keys:
            if thisKey=='q':  # it is equivalent to the string 'q'
                core.quit()
            else:
                print(thisKey.name, thisKey.tDown, thisKey.rt)
    """

    def __init__(self, code, tDown, name=None):
        self.code = code
        self.tDown = tDown
        self.duration = None
        self.rt = None
        if Keyboard._backend == 'event':  # we have event.getKeys()
            self.name = name
            self.rt = tDown
        elif Keyboard._backend == 'ptb':
            if code not in keyNames:
                self.name = 'n/a'
                logging.warning("Got keycode {} but that code isn't yet known")
            else:
                self.name = keyNames[code]
            if code not in keyNames:
                logging.warning('Keypress was given unknown key code ({})'.format(code))
                self.name = 'unknown'
            else:
                self.name = keyNames[code]
        elif Keyboard._backend == 'iohub':
            self.name = name

    def __eq__(self, other):
        return self.name == other

    def __ne__(self, other):
        return self.name != other


class _KeyBuffers(dict):
    """This ensures there is only one virtual buffer per physical keyboard.

    There is an option to get_event() from PTB without clearing but right
    now we are clearing when we poll so we need to make sure we have a single
    virtual buffer."""

    def getBuffer(self, kb_id, bufferSize=defaultBufferSize):
        if kb_id not in self:
            try:
                self[kb_id] = _KeyBuffer(bufferSize=bufferSize,
                                         kb_id=kb_id)
            except FileNotFoundError as e:
                if sys.platform == 'darwin':
                    # this is caused by a problem with SysPrefs
                    raise OSError("Failed to connect to Keyboard globally. "
                                  "You need to add PsychoPy App bundle (or the "
                                  "terminal if you run from terminal) to the "
                                  "System Preferences/Privacy/Accessibility "
                                  "(macOS <= 10.14) or "
                                  "System Preferences/Privacy/InputMonitoring "
                                  "(macOS >= 10.15).")
                else:
                    raise (e)

        return self[kb_id]


class _KeyBuffer(object):
    """This is our own local buffer of events with more control over clearing.

    The user shouldn't use this directly. It is fetched from the _keybuffers

    It stores events from a single physical device

    It's built on a collections.deque which is like a more efficient list
    that also supports a max length
    """

    def __init__(self, bufferSize, kb_id):
        self.bufferSize = bufferSize
        self._evts = deque()

        # create the PTB keyboard object and corresponding queue
        allInds, names, keyboards = hid.get_keyboard_indices()

        self._keys = deque()
        self._keysStillDown = deque()

        if kb_id == -1:
            self.dev = hid.Keyboard()  # a PTB keyboard object
        else:
            self.dev = hid.Keyboard(kb_id)  # a PTB keyboard object
        self.dev._create_queue(bufferSize, win_handle=None)

    def flush(self):
        """Flushes and processes events from the device to this software buffer
        """
        self._processEvts()

    def _flushEvts(self):
        while self.dev.flush(flush_type=_ptb_flush_type):
            evt, remaining = self.dev.queue_get_event()
            key = {}
            key['keycode'] = int(evt['Keycode'])
            key['down'] = bool(evt['Pressed'])
            key['time'] = evt['Time']
            self._evts.append(key)

    def getKeys(self, keyList=[], waitRelease=True, clear=True):
        """Return the KeyPress objects from the software buffer

        Parameters
        ----------
        keyList : list of key(name)s of interest
        waitRelease : if True then only process keys that are also released
        clear : clear any keys (that have been returned in this call)

        Returns
        -------
        A deque (like a list) of keys
        """
        self._processEvts()
        # if no conditions then no need to loop through
        if not keyList and not waitRelease:
            keyPresses = list(self._keysStillDown)
            for k in list(self._keys):
                if k not in keyPresses:
                    keyPresses.append(k)
            if clear:
                self._keys = deque()
                self._keysStillDown = deque()
            return keyPresses

        # otherwise loop through and check each key
        keyPresses = deque()
        for keyPress in self._keys:
            if waitRelease and not keyPress.duration:
                continue
            if keyList and keyPress.name not in keyList:
                continue
            keyPresses.append(keyPress)

        # clear keys in a second step (not during iteration)
        if clear:
            for key in keyPresses:
                self._keys.remove(key)

        return keyPresses

    def _clearEvents(self):
        self._evts.clear()

    def start(self):
        self.dev.queue_start()

    def stop(self):
        self.dev.queue_stop()

    def _processEvts(self):
        """Take a list of events and convert to a list of keyPresses with
        tDown and duration"""
        self._flushEvts()
        evts = deque(self._evts)
        self._clearEvents()
        for evt in evts:
            if evt['down']:
                newKey = KeyPress(code=evt['keycode'], tDown=evt['time'])
                self._keys.append(newKey)
                self._keysStillDown.append(newKey)
            else:
                for key in self._keysStillDown:
                    if key.code == evt['keycode']:
                        key.duration = evt['time'] - key.tDown
                        self._keysStillDown.remove(key)
                        break  # this key is done
                    else:
                        # we found a key that was first pressed before reading
                        pass


_keyBuffers = _KeyBuffers()

keyNamesWin = {
    49: '1', 50: '2', 51: '3', 52: '4', 53: '5',
    54: '6', 55: '7', 56: '8', 57: '9', 48: '0',
    65: 'a', 66: 'b', 67: 'c', 68: 'd', 69: 'e', 70: 'f',
    71: 'g', 72: 'h', 73: 'i', 74: 'j', 75: 'k', 76: 'l',
    77: 'm', 78: 'n', 79: 'o', 80: 'p', 81: 'q', 82: 'r',
    83: 's', 84: 't', 85: 'u', 86: 'v', 87: 'w', 88: 'x',
    89: 'y', 90: 'z',
    97: 'num_1', 98: 'num_2', 99: 'num_3',
    100: 'num_4', 101: 'num_5', 102: 'num_6', 103: 'num_7',
    104: 'num_8', 105: 'num_9', 96: 'num_0',
    112: 'f1', 113: 'f2', 114: 'f3', 115: 'f4', 116: 'f5',
    117: 'f6', 118: 'f7', 119: 'f8', 120: 'f9', 121: 'f10',
    122: 'f11', 123: 'f12',
    145: 'scrolllock', 19: 'pause', 36: 'home', 35: 'end',
    45: 'insert', 33: 'pageup', 46: 'delete', 34: 'pagedown',
    37: 'left', 40: 'down', 38: 'up', 39: 'right', 27: 'escape',
    144: 'numlock', 111: 'num_divide', 106: 'num_multiply',
    8: 'backspace', 109: 'num_subtract', 107: 'num_add',
    13: 'return', 222: 'pound', 161: 'lshift', 163: 'rctrl',
    92: 'rwindows', 32: 'space', 164: 'lalt', 165: 'ralt',
    91: 'lwindows', 93: 'menu', 162: 'lctrl', 160: 'lshift',
    20: 'capslock', 9: 'tab', 223: 'quoteleft', 220: 'backslash',
    188: 'comma', 190: 'period', 191: 'slash', 186: 'semicolon',
    192: 'apostrophe', 219: 'bracketleft', 221: 'bracketright',
    189: 'minus', 187: 'equal'
}

keyNamesMac = {
    4: 'a', 5: 'b', 6: 'c', 7: 'd', 8: 'e', 9: 'f', 10: 'g', 11: 'h', 12: 'i',
    13: 'j', 14: 'k', 15: 'l', 16: 'm', 17: 'n', 18: 'o', 19: 'p', 20: 'q',
    21: 'r', 22: 's', 23: 't', 24: 'u', 25: 'v', 26: 'w', 27: 'x', 28: 'y',
    29: 'z',
    30: '1', 31: '2', 32: '3', 33: '4', 34: '5', 35: '6', 36: '7',
    37: '8', 38: '9', 39: '0',
    40: 'return', 41: 'escape', 42: 'backspace', 43: 'tab', 44: 'space',
    45: 'minus', 46: 'equal',
    47: 'bracketleft', 48: 'bracketright', 49: 'backslash', 51: 'semicolon',
    52: 'apostrophe', 53: 'grave', 54: 'comma', 55: 'period', 56: 'slash',
    57: 'lshift',
    58: 'f1', 59: 'f2', 60: 'f3', 61: 'f4', 62: 'f5', 63: 'f6', 64: 'f7',
    65: 'f8', 66: 'f9', 67: 'f10', 68: 'f11', 69: 'f12',
    104: 'f13', 105: 'f14', 106: 'f15',
    107: 'f16', 108: 'f17', 109: 'f18', 110: 'f19',
    79: 'right', 80: 'left', 81: 'down', 82: 'up',
    224: 'lctrl', 225: 'lshift', 226: 'loption', 227: 'lcommand',
    100: 'function', 229: 'rshift', 230: 'roption', 231: 'rcommand',
    83: 'numlock', 103: 'num_equal', 84: 'num_divide', 85: 'num_multiply',
    86: 'num_subtract', 87: 'num_add', 88: 'num_enter', 99: 'num_decimal',
    98: 'num_0', 89: 'num_1', 90: 'num_2', 91: 'num_3', 92: 'num_4',
    93: 'num_5', 94: 'num_6', 95: 'num_7', 96: 'num_8', 97: 'num_9',
    74: 'home', 75: 'pageup', 76: 'delete', 77: 'end', 78: 'pagedown',
}

keyNamesLinux = {
    66: 'space', 68: 'f1', 69: 'f2', 70: 'f3', 71: 'f4', 72: 'f5',
    73: 'f6', 74: 'f7', 75: 'f8', 76: 'f9', 77: 'f10', 96: 'f11', 97: 'f12',
    79: 'scrolllock', 153: 'scrolllock', 128: 'pause', 119: 'insert', 111: 'home',
    120: 'delete', 116: 'end', 113: 'pageup', 118: 'pagedown', 136: 'menu', 112: 'up',
    114: 'left', 117: 'down', 115: 'right', 50: 'quoteleft',
    11: '1', 12: '2', 13: '3', 14: '4', 15: '5', 16: '6', 17: '7', 18: '8', 19: '9', 20: '0',
    21: 'minus', 22: 'equal', 23: 'backspace', 24: 'tab', 25: 'q', 26: 'w', 27: 'e', 28: 'r',
    29: 't', 30: 'y', 31: 'u', 32: 'i', 33: 'o', 34: 'p', 35: 'bracketleft', 36: 'bracketright',
    37: 'return', 67: 'capslock', 39: 'a', 40: 's', 41: 'd', 42: 'f', 43: 'g', 44: 'h', 45: 'j',
    46: 'k', 47: 'l', 48: 'semicolon', 49: 'apostrophe', 52: 'backslash', 51: 'lshift',
    95: 'less', 53: 'z', 54: 'x', 55: 'c', 56: 'v', 57: 'b', 58: 'n', 59: 'm',
    60: 'comma', 61: 'period', 62: 'slash', 63: 'rshift', 38: 'lctrl', 65: 'lalt',
    109: 'ralt', 106: 'rctrl', 78: 'numlock', 107: 'num_divide', 64: 'num_multiply',
    83: 'num_subtract', 80: 'num_7', 81: 'num_8', 82: 'num_9', 87: 'num_add', 84: 'num_4',
    85: 'num_5', 86: 'num_6', 88: 'num_1', 89: 'num_2', 90: 'num_3',
    105: 'num_enter', 91: 'num_0', 92: 'num_decimal', 10: 'escape'
}

if sys.platform == 'darwin':
    keyNames = keyNamesMac
elif sys.platform == 'win32':
    keyNames = keyNamesWin
else:
    keyNames = keyNamesLinux



"""Test PsychoPy sound.py using pyo backend
"""
import pytest
import shutil, os
from tempfile import mkdtemp
import numpy as np

from psychopy import prefs, core
from psychopy.tests import utils
from psychopy import sound

from importlib import reload

origSoundPref = prefs.hardware['audioLib']

# py.test --cov-report term-missing --cov sound.py tests/test_sound/test_sound_pyo.py


@pytest.mark.needs_sound
class TestSoundDevice():
    @classmethod
    def setup_class(self):
        self.contextName='sounddevice'
        prefs.hardware['audioLib'] = ['sounddevice']
        reload(sound)
        self.tmp = mkdtemp(prefix='psychopy-tests-sound')

        self.testFile = os.path.join(utils.TESTS_DATA_PATH,
                                     'Electronic_Chime-KevanGC-495939803.wav')

    @classmethod
    def teardown_class(self):
        prefs.hardware['audioLib'] = origSoundPref
        if hasattr(self, 'tmp'):
            shutil.rmtree(self.tmp, ignore_errors=True)

    def test_init(self):
        for note in ['A', 440, '440', [1,2,3,4], np.array([1,2,3,4])]:
            sound.Sound(note, secs=.1)
        with pytest.raises(ValueError):
            sound.Sound('this is not a file name')
        with pytest.raises(ValueError):
            sound.Sound(-1)  # negative frequency makes no sense

        points = 100
        snd = np.ones(points) / 20  # noqa

        s = sound.Sound(self.testFile)  # noqa

    def test_play(self):
        s = sound.Sound(secs=0.1)
        s.play()
        core.wait(s.getDuration()+.1)
        s.play(loops=1)  # exactly one loop
        core.wait(s.getDuration()*2+.1)  # allows coverage of _onEOS
        s.play(loops=-1)  # infinite loops
        s.stop()

    def test_start_stop(self):
        """only relevant for sound from files"""
        s1 = sound.Sound(self.testFile, startTime=0.5, stopTime=1.5)
        assert s1.getDuration() == 1
        s2 = sound.Sound(self.testFile, startTime=0.5)
        s3 = sound.Sound(self.testFile)
        assert s3.getDuration() > s2.getDuration() > s1.getDuration()
        s4 = sound.Sound(self.testFile, startTime=-1, stopTime=10000)
        assert s4.getDuration() == s3.getDuration()

    def test_methods(self):
        s = sound.Sound(secs=0.1)
        v = s.getVolume()
        assert v == 1
        s.setVolume(0.5)
        assert s.getVolume() == 0.5
        s.setLoops(2)
        assert s.getLoops() == 2

    def test_reinit_pyo(self):
        pytest.skip()
        # was stalling on some machines; revisit if decide to stick with pyo
        sound.initPyo()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""fORP fibre optic (MR-compatible) response devices by CurrentDesigns:
http://www.curdes.com/
This class is only useful when the fORP is connected via the serial port.

If you're connecting via USB, just treat it like a standard keyboard.
E.g., use a Keyboard component, and typically listen for Allowed keys
``'1', '2', '3', '4', '5'``. Or use ``event.getKeys()``.
"""

# Jeremy Gray and Dan Grupe developed the asKeys and baud parameters

from psychopy import logging, event
import sys
from collections import defaultdict

try:
    import serial
except ImportError:
    serial = False

BUTTON_BLUE = 1
BUTTON_YELLOW = 2
BUTTON_GREEN = 3
BUTTON_RED = 4
BUTTON_TRIGGER = 5
# Maps bit patterns to character codes
BUTTON_MAP = [
    (0x01, BUTTON_BLUE),
    (0x02, BUTTON_YELLOW),
    (0x04, BUTTON_GREEN),
    (0x08, BUTTON_RED),
    (0x10, BUTTON_TRIGGER)]


class ButtonBox:
    """Serial line interface to the fORP MRI response box.

    To use this object class, select the box use setting `serialPort`,
    and connect the serial line. To emulate key presses with a serial
    connection, use `getEvents(asKeys=True)` (e.g., to be able to use
    a RatingScale object during scanning). Alternatively connect the USB
    cable and use fORP to emulate a keyboard.

    fORP sends characters at 800Hz, so you should check the buffer
    frequently. Also note that the trigger event numpy the fORP is
    typically extremely short (occurs for a single 800Hz epoch).
    """

    def __init__(self, serialPort=1, baudrate=19200):
        """
        :Parameters:

            `serialPort` :
                should be a number (where 1=COM1, ...)
            `baud` :
                the communication rate (baud), eg, 57600
        """
        super(ButtonBox, self).__init__()
        if not serial:
            raise ImportError("The module serial is needed to connect to "
                              "fORP. On most systems this can be installed "
                              "with\n\t easy_install pyserial")

        self.port = serial.Serial(serialPort - 1, baudrate=baudrate,
                                  bytesize=8, parity='N', stopbits=1,
                                  timeout=0.001)
        if not self.port.isOpen():
            self.port.open()

        self.buttonStatus = defaultdict(bool)  # Defaults to False
        self.rawEvts = []
        self.pressEvents = []

    def clearBuffer(self):
        """Empty the input buffer of all characters"""
        self.port.flushInput()

    def clearStatus(self):
        """ Resets the pressed statuses, so getEvents will return pressed
        buttons, even if they were already pressed in the last call.
        """
        for k in self.buttonStatus:
            self.buttonStatus[k] = False

    def getEvents(self, returnRaw=False, asKeys=False, allowRepeats=False):
        """Returns a list of unique events (one event per button pressed)
        and also stores a copy of the full list of events since last
        getEvents() (stored as ForpBox.rawEvts)

        `returnRaw` :
            return (not just store) the full event list

        `asKeys` :
            If True, will also emulate pyglet keyboard events, so that
            button 1 will register as a keyboard event with value "1",
            and as such will be detectable using `event.getKeys()`

        `allowRepeats` :
            If True, this will return pressed buttons even if they were held
            down between calls to getEvents(). If the fORP is on the "Eprime"
            setting, you will get a stream of button presses while a button is
            held down. On the "Bitwise" setting, you will get a set of all
            currently pressed buttons every time a button is pressed or
            released.
            This option might be useful if you think your participant may be
            holding the button down before you start checking for presses.
        """
        nToGet = self.port.inWaiting()
        evtStr = self.port.read(nToGet)
        self.rawEvts = []
        self.pressEvents = []
        if allowRepeats:
            self.clearStatus()
        # for each character convert to an ordinal int value (numpy the ascii
        # chr)
        for thisChr in evtStr:
            pressCode = ord(thisChr)
            self.rawEvts.append(pressCode)
            decodedEvents = self._generateEvents(pressCode)
            self.pressEvents += decodedEvents
            if asKeys:
                for code in decodedEvents:
                    event._onPygletKey(symbol=code, modifiers=0)
                    # better as: emulated='fORP_bbox_asKey', but need to
                    # adjust event._onPygletKey and the symbol conversion
                    # pyglet.window.key.symbol_string(symbol).lower()
        # return the abbreviated list if necessary
        if returnRaw:
            return self.rawEvts
        else:
            return self.getUniqueEvents()

    def _generateEvents(self, pressCode):
        """For a given button press, returns a list buttons that went from
        unpressed to pressed.
        Also flags any unpressed buttons as unpressed.

        `pressCode` :
            a number with a bit set for every button currently pressed.
        """

        curStatuses = self.__class__._decodePress(pressCode)
        pressEvents = []
        for button, pressed in curStatuses:
            if pressed and not self.buttonStatus[button]:
                # We're transitioning to pressed...
                pressEvents.append(button)
                self.buttonStatus[button] = True
            if not pressed:
                self.buttonStatus[button] = False
        return pressEvents

    @classmethod
    def _decodePress(kls, pressCode):
        """Returns a list of buttons and whether they're pressed, given a
            character code.

        `pressCode` :
            A number with a bit set for every button currently pressed. Will
            be between 0 and 31.
        """

        return [(mapping[1], bool(mapping[0] & pressCode))
                for mapping in BUTTON_MAP]

    def getUniqueEvents(self, fullEvts=False):
        """Returns a Python set of the unique (unordered) events of either
        a list given or the current rawEvts buffer
        """
        if fullEvts:
            return set(self.rawEvts)
        return set(self.pressEvents)



# pyxhook -- an extension to emulate some of the PyHook library on linux.
#
# Copyright (C) 2008 Tim Alexander <dragonfyre13@gmail.com>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Thanks to Alex Badea <vamposdecampos@gmail.com> for writing the Record
#    demo for the xlib libraries. It helped me immensely working with these
#    in this library.
#
#    Thanks to the python-xlib team. This wouldn't have been possible without
#    your code.
#
#    This requires:
#    at least python-xlib 1.4
#    xwindows must have the "record" extension present, and active.
#
#    This file has now been somewhat extensively modified by
#    Daniel Folkinshteyn <nanotube@users.sf.net>
#    So if there are any bugs, they are probably my fault. :)
#

#   January 2013: File modified by
#      Sol Simpson (sol@isolver-software.com), with some cleanup done and
#      modifications made so it integrated with the ioHub module more effecively
#     ( but therefore making this version not useful for general application usage)
#
#   March, 2013: -Fixed an existing bug that caused capslock not to have an effect,
#              -Added tracking of what keys are pressed and how many auto repeat
#              press events each has received.
#   April, 2013: - Modified to directly return ioHub device event arrays
#             - optimized keysym lookup by loading into a dict cache
#             - started adding support for reporting unicode keys

import threading
import unicodedata
import ctypes as ct
from Xlib import X, display
from Xlib.ext import record
from Xlib.protocol import rq
from . import xlib as _xlib

from .computer import Computer
from ..constants import EventConstants, MouseConstants, ModifierKeyCodes
from ..errors import print2err

jdumps = lambda x: str(x)
try:
    import ujson
    jdumps = ujson.dumps
except Exception:
    import json
    jdumps = json.dumps

getTime = Computer.getTime

#######################################################################
########################START CLASS DEF################################
#######################################################################


def event2json(event):
    """
    Instance Variable: KeyButtonPointerEvent time
        The server X time when this event was generated.
    Instance Variable: KeyButtonPointerEvent root
        The root window which the source window is an inferior of.
    Instance Variable: KeyButtonPointerEvent window
        The window the event is reported on.
    Instance Variable: KeyButtonPointerEvent same_screen
        Set to 1 if window is on the same screen as root, 0 otherwise.
    Instance Variable: KeyButtonPointerEvent child
        If the source window is an inferior of window, child is set to the child of window that is the ancestor of (or is) the source window. Otherwise it is set to X.NONE.
    Instance Variable: KeyButtonPointerEvent root_x
        Instance Variable: KeyButtonPointerEvent root_y
    The pointer coordinates at the time of the event, relative to the root window.
        Instance Variable: KeyButtonPointerEvent event_x
    Instance Variable: KeyButtonPointerEvent event_y
        The pointer coordinates at the time of the event, relative to window. If window is not on the same screen as root, these are set to 0.
    Instance Variable: KeyButtonPointerEvent state
        The logical state of the button and modifier keys just before the event.
    Instance Variable: KeyButtonPointerEvent detail
        For KeyPress and KeyRelease, this is the keycode of the event key.
        For ButtonPress and ButtonRelease, this is the button of the event.
        For MotionNotify, this is either X.NotifyNormal or X.NotifyHint.
    """
    return jdumps(dict(type=event.type,
                       send_event=event.send_event,
                       time=event.time,
                       root=str(event.root),
                       window=str(event.window),
                       same_screen=event.same_screen,
                       child=str(event.child),
                       root_x=event.root_x,
                       root_y=event.root_y,
                       event_x=event.event_x,
                       event_y=event.event_y,
                       state=event.state,
                       detail=event.detail))

# xlib modifier constants to iohub str constants
key_mappings = {'num_lock': 'numlock',
                'caps_lock': 'capslock',
                'scroll_lock': 'scrolllock',
                'shift_l': 'lshift',
                'shift_r': 'rshift',
                'alt_l': 'lalt',
                'alt_r': 'ralt',
                'control_l': 'lctrl',
                'control_r': 'rctrl',
                'super_l': 'lcmd',
                'super_r': 'rcmd'
                }


class HookManager(threading.Thread):
    """Creates a separate thread that starts the Xlib Record functionality,
    capturing keyboard and mouse events and transmitting them to the associated
    callback functions set."""
    DEVICE_TIME_TO_SECONDS = 0.001
    evt_types = [
        X.KeyRelease,
        X.KeyPress,
        X.ButtonRelease,
        X.ButtonPress,
        X.MotionNotify]

    def __init__(self, log_event_details=False):
        threading.Thread.__init__(self)
        self.finished = threading.Event()
        self._running = False
        self.log_events = log_event_details
        self.log_events_file = None

        # Give these some initial values
        self.mouse_position_x = 0
        self.mouse_position_y = 0

        # Assign default function actions (do nothing).
        self.KeyDown = lambda x: True
        self.KeyUp = lambda x: True
        self.MouseAllButtonsDown = lambda x: True
        self.MouseAllButtonsUp = lambda x: True
        self.MouseAllMotion = lambda x: True
        self.contextEventMask = [X.KeyPress, X.MotionNotify]

        # Used to hold any keys currently pressed and the repeat count
        # of each key.
        self.key_states = dict()

        self.contextEventMask = [X.KeyPress, X.MotionNotify]

        # Hook to our display.
        self.local_dpy = display.Display()
        self.record_dpy = display.Display()

        self.ioHubMouseButtonMapping = {1: 'MOUSE_BUTTON_LEFT',
                                        2: 'MOUSE_BUTTON_MIDDLE',
                                        3: 'MOUSE_BUTTON_RIGHT'
                                        }
        self.pressedMouseButtons = 0
        self.scroll_y = 0

        # Direct xlib ctypes wrapping for better / faster keyboard event -> key,
        # char field mapping.
        self._xlib = _xlib
        self._xdisplay = _xlib.XOpenDisplay(None)
        self._xroot = _xlib.XDefaultRootWindow(self._xdisplay)
        self._keysym = _xlib.KeySym()
        self._compose = _xlib.XComposeStatus()
        self._tmp_compose = _xlib.XComposeStatus()
        self._revert = ct.c_int(0)
        self._charbuf = (ct.c_char * 17)()
        self._cwin = _xlib.Window()
        self._revert_to_return = ct.c_int()

        self._xkey_evt = _xlib.XKeyEvent()
        self._xkey_evt.serial = 1  # not known, make it up.
        self._xkey_evt.send_event = 0
        self._xkey_evt.subwindow = 0
        self._xkey_evt.display = self._xdisplay
        self._xkey_evt.root = self._xroot  # ', Window),
        self._xkey_evt.subwindow = 0  # ', Window)

    def run(self):
        self._running = True
        # Check if the extension is present
        if not self.record_dpy.has_extension('RECORD'):
            print2err(
                'RECORD extension not found. ioHub can not use python Xlib. Exiting....')
            return False

        # Create a recording context; we only want key and mouse events
        self.ctx = self.record_dpy.record_create_context(
            0,
            [record.AllClients],
            [{
                'core_requests': (0, 0),
                'core_replies': (0, 0),
                'ext_requests': (0, 0, 0, 0),
                'ext_replies': (0, 0, 0, 0),
                'delivered_events': (0, 0),
                # (X.KeyPress, X.ButtonPress),
                'device_events': tuple(self.contextEventMask),
                'errors': (0, 0),
                'client_started': False,
                'client_died': False,
            }])

        if self.log_events:
            import datetime

            cdate = datetime.datetime.now().strftime('%Y_%m_%d_%H_%M')
            with open('x11_events_{0}.log'.format(cdate), 'w') as self.log_events_file:
                # Enable the context; this only returns after a call to record_disable_context,
                # while calling the callback function in the meantime
                self.record_dpy.record_enable_context(
                    self.ctx, self.processevents)
                # Finally free the context
                self.record_dpy.record_free_context(self.ctx)
        else:
            self.record_dpy.record_enable_context(self.ctx, self.processevents)
            # Finally free the context
            self.record_dpy.record_free_context(self.ctx)

    def cancel(self):
        self.finished.set()
        self._running = False
        self.local_dpy.record_disable_context(self.ctx)
        self.local_dpy.flush()
        try:
            self._xlib.XCloseDisplay(self._xdisplay)
            self._xlib = None
        except AttributeError:
            pass
        
    def printevent(self, event):
        print2err(event)

    def HookKeyboard(self):
        pass

    def HookMouse(self):
        pass

    def updateKeysPressedState(self, key_str, pressed_event):
        keyautocount = self.key_states.setdefault(key_str, -1)
        if pressed_event:
            self.key_states[key_str] = keyautocount + 1
        else:
            del self.key_states[key_str]
        return self.key_states

    def isKeyPressed(self, key_str_id):
        """Returns 0 if key is not pressed, otherwise a.

        positive int, representing the auto repeat count ( return val - 1)
        of key press events that have occurred for the key.

        """
        return self.key_states.get(key_str_id, 0)

    def getPressedKeys(self, repeatCounts=False):
        """
        If repeatCounts == False (default), returns a list
        of all the key symbol strings currently pressed.

        If repeatCounts == True, returns the dict of key
        sybol strs, pressedCount.
        """
        if repeatCounts:
            return self.key_states
        return list(self.key_states.keys())

    def processevents(self, reply):
        logged_time = getTime()
        if reply.category != record.FromServer:
            return
        if reply.client_swapped:
            print2err(
                'pyXlib: * received swapped protocol data, cowardly ignored')
            return
        if not len(reply.data):# or ord(reply.data[0]) < 2:
            # not an event
            return
        data = reply.data
        while len(data):
            event, data = rq.EventField(None).parse_binary_value(
                data, self.record_dpy.display, None, None)

            if self.log_events_file and event.type in self.evt_types:
                self.log_events_file.write(event2json(event) + '\n')

            event.iohub_logged_time = logged_time
            if event.type == X.KeyPress:
                hookevent = self.makekeyhookevent(event)
                self.KeyDown(hookevent)
            elif event.type == X.KeyRelease:
                hookevent = self.makekeyhookevent(event)
                self.KeyUp(hookevent)
            elif event.type == X.ButtonPress:
                hookevent = self.buttonpressevent(event)
                self.MouseAllButtonsDown(hookevent)
            elif event.type == X.ButtonRelease and event.detail not in (4, 5):
                # 1 mouse wheel scroll event was generating a button press
                # and a button release event for each single scroll, so allow
                # wheel scroll events through for buttonpressevent, but not for
                # buttonreleaseevent so 1 scroll action causes 1 scroll event.
                hookevent = self.buttonreleaseevent(event)
                self.MouseAllButtonsUp(hookevent)
            elif event.type == X.MotionNotify:
                # use mouse moves to record mouse position, since press and release events
                # do not give mouse position info (event.root_x and event.root_y have
                # bogus info).
                hookevent = self.mousemoveevent(event)
                self.MouseAllMotion(hookevent)

    def buttonpressevent(self, event):
        r = self.makemousehookevent(event)
        return r

    def buttonreleaseevent(self, event):
        r = self.makemousehookevent(event)
        return r

    def mousemoveevent(self, event):
        self.mouse_position_x = event.root_x
        self.mouse_position_y = event.root_y
        r = self.makemousehookevent(event)
        return r

    def getKeyChar(self, kb_event):
        keycode = kb_event.detail

        # get char string for keyboard event.
        _xlib.XGetInputFocus(
            self._xdisplay, ct.byref(
                self._cwin), ct.byref(
                self._revert_to_return))
        self._xkey_evt.window = self._cwin
        self._xkey_evt.type = kb_event.type
        # >>> How many of these event fields really need to be filled in for XLookupString to work?
        self._xkey_evt.time = kb_event.time  # ', Time),
        self._xkey_evt.x = kb_event.event_x  # ', c_int),
        self._xkey_evt.y = kb_event.event_y  # ', c_int),
        self._xkey_evt.x_root = kb_event.root_x  # ', c_int),
        self._xkey_evt.y_root = kb_event.root_y  # ', c_int),
        self._xkey_evt.state = kb_event.state  # ', c_uint),
        self._xkey_evt.keycode = keycode  # ', c_uint),
        self._xkey_evt.same_screen = kb_event.same_screen  # ', c_int),
        # <<<

        count = _xlib.XLookupString(
            ct.byref(
                self._xkey_evt), self._charbuf, 16, ct.byref(
                self._keysym), ct.byref(
                self._compose))
        char = ''
        ucat = ''
        if count > 0:
            char = u'' + self._charbuf[0:count].decode('utf-8')
            ucat = unicodedata.category(char)
            char = char.encode('utf-8')

        # special char char handling
        # Setting count == 0 makes key use XKeysymToString return value.
        if ucat.lower() == 'cc':
            if char == '\r':
                char = '\n'
                count = 0
            elif char == '\t':
                count = 0
            elif char != '\n':
                char = ''
                count = 0

        # Get key value
        keysym = _xlib.XKeycodeToKeysym(self._xdisplay, keycode, 0)
        key = _xlib.XKeysymToString(keysym)
        if isinstance(key, bytes):
            key = key.decode('utf-8')
        if isinstance(char, bytes):
            char = char.decode('utf-8')
            
        if key:
            key = key.lower()
            if key and key.startswith('kp_'):
                key = 'num_%s' % (key[3:])
            elif key in key_mappings:
                key = key_mappings[key]
            elif count > 0:
                self._xkey_evt.state = 0
                count = _xlib.XLookupString(
                    ct.byref(
                        self._xkey_evt), self._charbuf, 16, ct.byref(
                        self._keysym), ct.byref(
                        self._tmp_compose))
                key = ''
                if count > 0:
                    key = u'' + self._charbuf[0:count].decode('utf-8')
                    key = key.encode('utf-8')
        else:
            key = ''
        return keycode, keysym, key, char

    def makekeyhookevent(self, event):
        """Creates a ioHub keyboard event in list format, completing as much as
        possible from within pyXHook."""
        key_code, keysym, key, char = self.getKeyChar(event)

        event_type_id = EventConstants.KEYBOARD_PRESS
        is_pressed_key = event.type == X.KeyPress
        if not is_pressed_key:
            is_pressed_key = False
            event_type_id = EventConstants.KEYBOARD_RELEASE

        pressed_keys = self.updateKeysPressedState(key, is_pressed_key)
        auto_repeat_count = pressed_keys.get(key, 0)

        mod_mask = event.state
        modifier_key_state = 0
        # Update currently active modifiers (modifier_key_state)
        #
        if mod_mask & 2 == 2:
            # capslock is active:
            modifier_key_state += ModifierKeyCodes.capslock
        if mod_mask & 16 == 16:
            # numlock is active:
            modifier_key_state += ModifierKeyCodes.numlock
        for pk in pressed_keys:
            if pk not in ['capslock', 'numlock']:
                is_mod_id = ModifierKeyCodes.getID(pk)
                if is_mod_id:
                    modifier_key_state += is_mod_id

        # return event to iohub
        return [[0,
                 0,
                 0,  # device id (not currently used)
                 0,  # to be assigned by ioHub server# Device._getNextEventID(),
                 event_type_id,
                 event.time * self.DEVICE_TIME_TO_SECONDS,
                 event.iohub_logged_time,
                 event.iohub_logged_time,
                 0.0,
                 # confidence interval not set for keyboard or mouse devices.
                 0.0,  # delay not set for keyboard or mouse devices.
                 0,  # filter level not used
                 auto_repeat_count,  # auto_repeat
                 key_code,  # scan / Keycode of event.
                 keysym,  # KeyID / VK code for key pressed
                 0,  # unicode value for char, otherwise, 0
                 key,  # psychpy key event val
                 modifier_key_state,
                 # The logical state of the button and modifier keys just
                 # before the event.
                 int(self._cwin.value),
                 char,
                 # utf-8 encoded char or label for the key. (depending on
                 # whether it is a visible char or not)
                 0.0,
                 0
                 ], ]

    def makemousehookevent(self, event):
        """Creates an incomplete ioHub keyboard event in list format. It is
        incomplete as some of the elements of the array are filled in by the
        ioHub server when it receives the events.

        For event attributes see: http://python-xlib.sourceforge.net/doc/html/python-xlib_13.html

        time
        The server X time when this event was generated.

        root
        The root window which the source window is an inferior of.

        window
        The window the event is reported on.

        same_screen
        Set to 1 if window is on the same screen as root, 0 otherwise.

        child
        If the source window is an inferior of window, child is set to the child of window that is the ancestor of (or is) the source window. Otherwise it is set to X.NONE.

        root_x
        root_y
        The pointer coordinates at the time of the event, relative to the root window.

        event_x
        event_y
        The pointer coordinates at the time of the event, relative to window. If window is not on the same screen as root, these are set to 0.

        state
        The logical state of the button and modifier keys just before the event.

        detail
        For KeyPress and KeyRelease, this is the keycode of the event key.
        For ButtonPress and ButtonRelease, this is the button of the event.
        For MotionNotify, this is either X.NotifyNormal or X.NotifyHint.

        """
        px, py = event.root_x, event.root_y
        event_type_id = 0
        event_state = []
        event_detail = []
        dy = 0

        _xlib.XGetInputFocus(
            self._xdisplay, ct.byref(
                self._cwin), ct.byref(
                self._revert_to_return))

        if event.type == 6:
            if event.state < 128:
                event_type_id = EventConstants.MOUSE_MOVE
            else:
                event_type_id = EventConstants.MOUSE_DRAG

        if event.type in [4, 5]:
            if event.type == 5:
                event_type_id = EventConstants.MOUSE_BUTTON_RELEASE
            elif event.type == 4:
                event_type_id = EventConstants.MOUSE_BUTTON_PRESS

            if event.detail == 4 and event.type == 4:
                event_type_id = EventConstants.MOUSE_SCROLL
                self.scroll_y += 1
                dy = 1
            elif event.detail == 5 and event.type == 4:
                event_type_id = EventConstants.MOUSE_SCROLL
                self.scroll_y -= 1
                dy = -1

        if event.state & 1 == 1:
            event_state.append('SHIFT')
        if event.state & 4 == 4:
            event_state.append('ALT')
        if event.state & 64 == 64:
            event_state.append('WIN_MENU')
        if event.state & 8 == 8:
            event_state.append('CTRL')

            event_state.append('MOUSE_BUTTON_LEFT')
        if event.state & 512 == 512:
            event_state.append('MOUSE_BUTTON_MIDDLE')
        if event.state & 1024 == 1024:
            event_state.append('MOUSE_BUTTON_RIGHT')

        if event.detail == 1:
            event_detail.append('MOUSE_BUTTON_LEFT')
        if event.detail == 2:
            event_detail.append('MOUSE_BUTTON_MIDDLE')
        if event.detail == 3:
            event_detail.append('MOUSE_BUTTON_RIGHT')

        # TODO implement mouse event to display index detection
        display_index = 0

        currentButton = 0
        pressed = 0
        currentButtonID = 0
        if event.type in [
                4,
                5] and event_type_id != EventConstants.MOUSE_SCROLL:

            currentButton = self.ioHubMouseButtonMapping.get(event.detail)
            currentButtonID = MouseConstants.getID(currentButton)

            pressed = event.type == 4

            if pressed is True:
                self.pressedMouseButtons += currentButtonID
            else:
                self.pressedMouseButtons -= currentButtonID

        return [[0,
                 0,
                 0,  # device id (not currently used)
                 0,  # to be assigned by ioHub server# Device._getNextEventID(),
                 event_type_id,
                 event.time * self.DEVICE_TIME_TO_SECONDS,
                 event.iohub_logged_time,
                 event.iohub_logged_time,
                 0.0,
                 # confidence interval not set for keyboard or mouse devices.
                 0.0,  # delay not set for keyboard or mouse devices.
                 0,  # filter level not used
                 display_index,  # event.DisplayIndex,
                 pressed,
                 currentButtonID,
                 self.pressedMouseButtons,
                 px,  # mouse x pos
                 py,  # mouse y post
                 0,  # scroll_dx not supported
                 0,  # scroll_x
                 dy,
                 self.scroll_y,
                 0,  # mod state, filled in when event received by iohub
                 int(self._cwin.value)], ]
        # TO DO: Implement multimonitor location based on mouse location support.
        # Currently always uses monitor index 0

#
#
#######################################################################



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import os
from .. import _DATA_STORE_AVAILABLE, IOHUB_DIRECTORY
from . import ioHubConnection
from ..util import yload, yLoader, readConfig
from psychopy import logging


def launchHubServer(**kwargs):
    """
    Starts the ioHub Server subprocess, and return a
    :class:`psychopy.iohub.client.ioHubConnection` object that is used to
    access enabled iohub device's events, get events, and control the
    ioHub process during the experiment.

    By default (no kwargs specified), the ioHub server does not create an
    ioHub HDF5 file, events are available to the experiment program at runtime.
    The following Devices are enabled by default:

    - Keyboard: named 'keyboard', with runtime event reporting enabled.
    - Mouse: named 'mouse', with runtime event reporting enabled.
    - Monitor: named 'monitor'.
    - Experiment: named 'experiment'.

    To customize how the ioHub Server is initialized when started, use
    one or more of the following keyword arguments when calling the function:

    Parameters
    -----------
    experiment_code : str, <= 256 char
        If experiment_code is provided, an ioHub HDF5 file will be created for the session.
    session_code : str, <= 256 char
        When specified, used as the name of the ioHub HDF5 file created for the session.
    experiment_info : dict
        Can be used to save the following experiment metadata fields:
        code (<=256 chars), title (<=256 chars), description (<=4096 chars), version (<=32 chars)
    session_info : dict
        Can be used to save the following session metadata fields:
        code (<=256 chars), name (<=256 chars), comments (<=4096 chars), user_variables (dict)
    datastore_name : str
        Used to provide an ioHub HDF5 file name different than the session_code.
    window : :class:`psychopy.visual.Window`
        The psychoPy experiment window being used. Information like display size, viewing distance,
        coord / color type is used to update the ioHub Display device.
    iohub_config_name : str
        Specifies the name of the iohub_config.yaml file that contains the ioHub Device
        list to be used by the ioHub Server. i.e. the 'device_list' section of the yaml file.
    iohub.device.path : str
        Add an ioHub Device by using the device class path as the key, and the device's configuration
        in a dict value.
    psychopy_monitor : (deprecated)
        The path to a Monitor Center config file

    Examples:

        A. Wait for the 'q' key to be pressed::

            from psychopy.iohub.client import launchHubServer

            # Start the ioHub process. 'io' can now be used during the
            # experiment to access iohub devices and read iohub device events.
            io=launchHubServer()

            print("Press any Key to Exit Example.....")

            # Wait until a keyboard event occurs
            keys = io.devices.keyboard.waitForKeys(keys=['q',])

            print("Key press detected: {}".format(keys))
            print("Exiting experiment....")

            # Stop the ioHub Server
            io.quit()

    Please see the psychopy/demos/coder/iohub/launchHub.py demo for examples
    of different ways to use the launchHubServer function.
    """
    experiment_code = kwargs.get('experiment_code', None)
    if experiment_code:
        del kwargs['experiment_code']
    experiment_info = kwargs.get('experiment_info')
    if experiment_info:
        del kwargs['experiment_info']
        for k, v in list(experiment_info.items()):
            if k in ['code', 'title', 'description', 'version']:
                experiment_info[k] = u"{}".format(v)
        if experiment_info.get('code'):
            experiment_code = experiment_info['code']
        elif experiment_code:
            experiment_info['code'] = experiment_code
    elif experiment_code:
        experiment_info = dict(code=experiment_code)

    session_code = kwargs.get('session_code', None)
    if session_code:
        del kwargs['session_code']
    session_info = kwargs.get('session_info')
    if session_info:
        del kwargs['session_info']
        for k, v in list(session_info.items()):
            if k in ['code', 'name', 'comments']:
                session_info[k] = u"{}".format(v)
            elif k == 'user_variables':
                session_info[k] = v
        if session_info.get('code'):
            session_code = session_info['code']
        elif session_code:
            session_info['code'] = session_code
    elif session_code:
        session_info = dict(code=session_code)
    else:        
        session_info = {}

    if experiment_code and not session_code:
        # this means we should auto_generate a session code
        import datetime
        dtstr = datetime.datetime.now().strftime('%d_%m_%Y_%H_%M')
        session_info['code'] = session_code = u"S_{0}".format(dtstr)

    datastore_name = None
    if _DATA_STORE_AVAILABLE is True:
        datastore_name = kwargs.get('datastore_name')
        if datastore_name:
            del kwargs['datastore_name']
        elif session_code:
            datastore_name = session_code

    monitor_devices_config = None
    iohub_conf_file_name = kwargs.get('iohub_config_name')
    if iohub_conf_file_name:
        # Load the specified iohub configuration file,
        # converting it to apython dict.
        with open(iohub_conf_file_name, 'r') as iohub_conf_file:
            _temp_conf_read = yload(iohub_conf_file, Loader=yLoader)
            monitor_devices_config = _temp_conf_read.get('monitor_devices')
            del kwargs['iohub_config_name']

    device_dict = {}
    if monitor_devices_config:
        device_dict = monitor_devices_config

    if isinstance(device_dict, (list, tuple)):
        tempdict_ = {}
        for ddict in device_dict:
            tempdict_[list(ddict.keys())[0]] = list(ddict.values())[0]
        device_dict = tempdict_

    # PsychoPy Window & Monitor integration

    # Get default iohub display config settings for experiment
    display_config = device_dict.get('Display', {})
    if display_config:
        del device_dict['Display']

    # Check for a psychopy_monitor_name name
    monitor_name = kwargs.get('psychopy_monitor_name', kwargs.get('monitor_name'))
    if monitor_name:
        if kwargs.get('psychopy_monitor_name'):
            del kwargs['psychopy_monitor_name']
        else:
            del kwargs['monitor_name']

    window = kwargs.get('window')
    if window:
        kwargs['window'] = None
        del kwargs['window']
        # PsychoPy Window has been provided, so read all info needed for iohub Display from Window
        if window.units:
            display_config['reporting_unit_type'] = window.units
        if window.colorSpace:
            display_config['color_space'] = window.colorSpace
        display_config['device_number'] = window.screen

        if window.monitor.name == "__blank__":
            logging.warning("launchHubServer: window.monitor.name is '__blank__'. "
                            "Create the PsychoPy window with a valid Monitor name.")
        elif window.monitor.name:
            display_config['psychopy_monitor_name'] = window.monitor.name
            display_config['override_using_psycho_settings'] = True

        if not window._isFullScr:
            logging.warning("launchHubServer: If using the iohub mouse or eyetracker devices, fullScr should be True.")

    elif monitor_name:
        display_config['psychopy_monitor_name'] = monitor_name
        display_config['override_using_psycho_settings'] = True
        logging.warning("launchHubServer: Use of psychopy_monitor_name is deprecated. "
                        "Please use window= and provide a psychopy window that has a .monitor.")

    device_dict.update(kwargs)
    device_list = []

    def isFunction(func):
        import types
        return isinstance(func, types.FunctionType)

    def func2str(func):
        return '%s.%s' % (func.__module__, func.__name__)

    def configfuncs2str(config):
        for key, val in list(config.items()):
            if isinstance(val, dict):
                configfuncs2str(val)
            if isFunction(val):
                config[key] = func2str(val)

    configfuncs2str(device_dict)

    # Add Display device as first in list of devices to be sent to iohub
    device_list.append(dict(Display=display_config))

    # Ensure a Experiment, Keyboard, and Mouse Devices have been defined.
    # If not, create them.
    check_for_devs = ['Experiment', 'Keyboard', 'Mouse']
    for adev_name in check_for_devs:
        if adev_name not in device_dict:
            device_list.append({adev_name: {}})
        else:
            device_list.append({adev_name: device_dict[adev_name]})
            del device_dict[adev_name]
    
    iohub_config = dict()
    def_ioconf = readConfig(os.path.join(IOHUB_DIRECTORY, u'default_config.yaml'))
    # Add remaining defined devices to the device list.
    for class_name, device_config in device_dict.items():
        if class_name in def_ioconf:
            # not a device, a top level iohub config param
            iohub_config[class_name] = device_config
        else:
            device_list.append({class_name: device_config})
            
    # Create an ioHub configuration dictionary.
    iohub_config['monitor_devices'] = device_list

    if _DATA_STORE_AVAILABLE and (datastore_name or session_code):
        # If datastore_name kwarg or experiment code has been provided,
        # then enable saving of device events to the iohub datastore hdf5 file.
        # If datastore_name kwarg was provided, it is used for the hdf5 file
        # name, otherwise the session code is used. This avoids different
        # experiments / sessions running in the same directory from using
        # the same hdf5 file name.
        if datastore_name is None:
            datastore_name = session_code
        parent_dir, datastore_name = os.path.split(datastore_name)
        iohub_config['data_store'] = dict(enable=True,
                                          filename=datastore_name,
                                          experiment_info=experiment_info,
                                          session_info=session_info)
        if parent_dir:
            iohub_config['data_store']['parent_dir'] = parent_dir

    return ioHubConnection(iohub_config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import os
import atexit
import numpy as np
from pkg_resources import parse_version
from ..server import DeviceEvent
from ..constants import EventConstants
from ..errors import ioHubError, printExceptionDetailsToStdErr, print2err

import tables
from tables import parameters, StringCol, UInt32Col, UInt16Col, NoSuchNodeError

if parse_version(tables.__version__) < parse_version('3'):
    from tables import openFile as open_file

    create_table = "createTable"
    create_group = "createGroup"
    f_get_child = "_f_getChild"
else:
    from tables import open_file

    create_table = "create_table"
    create_group = "create_group"
    _f_get_child = "_f_get_child"

parameters.MAX_NUMEXPR_THREADS = None
"""The maximum number of threads that PyTables should use internally in
Numexpr.  If `None`, it is automatically set to the number of cores in
your machine."""

parameters.MAX_BLOSC_THREADS = None
"""The maximum number of threads that PyTables should use internally in
Blosc.  If `None`, it is automatically set to the number of cores in
your machine."""

DATA_FILE_TITLE = "ioHub DataStore - Experiment Data File."
FILE_VERSION = '0.9.1.2'
SCHEMA_AUTHORS = 'Sol Simpson'
SCHEMA_MODIFIED_DATE = 'October 27, 2021'


class DataStoreFile():
    def __init__(self, fileName, folderPath, fmode='a', iohub_settings=None):
        self.fileName = fileName
        self.folderPath = folderPath
        self.filePath = os.path.join(folderPath, fileName)

        if iohub_settings.get('multiple_sessions', False) is False:
            fmode = 'w'

        self.settings = iohub_settings

        self.active_experiment_id = None
        self.active_session_id = None

        self.flushCounter = self.settings.get('flush_interval', 32)
        self._eventCounter = 0

        self.TABLES = dict()
        self._eventGroupMappings = dict()
        self.emrtFile = open_file(self.filePath, mode=fmode)

        atexit.register(close_open_data_files, False)

        if len(self.emrtFile.title) == 0:
            self.buildOutTemplate()
            self.flush()
        else:
            self.loadTableMappings()

    def loadTableMappings(self):
        # create meta-data tables
        self.TABLES['EXPERIMENT_METADETA'] = self.emrtFile.root.data_collection.experiment_meta_data
        self.TABLES['SESSION_METADETA'] = self.emrtFile.root.data_collection.session_meta_data
        self.TABLES['CLASS_TABLE_MAPPINGS'] = self.emrtFile.root.class_table_mapping

    def buildOutTemplate(self):
        self.emrtFile.title = DATA_FILE_TITLE
        self.emrtFile.FILE_VERSION = FILE_VERSION
        self.emrtFile.SCHEMA_DESIGNER = SCHEMA_AUTHORS
        self.emrtFile.SCHEMA_MODIFIED = SCHEMA_MODIFIED_DATE

        create_group_func = getattr(self.emrtFile, create_group)
        create_table_func = getattr(self.emrtFile, create_table)

        # CREATE GROUPS
        self.TABLES['CLASS_TABLE_MAPPINGS'] = create_table_func(self.emrtFile.root, 'class_table_mapping',
                                                                ClassTableMappings, title='ioHub DeviceEvent Class to '
                                                                                          'DataStore Table Mappings.')

        create_group_func(self.emrtFile.root, 'data_collection', title='Data collected using ioHub.')
        self.flush()

        create_group_func(self.emrtFile.root.data_collection, 'events', title='Events collected using ioHub.')

        create_group_func(self.emrtFile.root.data_collection, 'condition_variables', title="Experiment Condition "
                                                                                           "Variable Data.")
        self.flush()

        self.TABLES['EXPERIMENT_METADETA'] = create_table_func(self.emrtFile.root.data_collection,
                                                               'experiment_meta_data', ExperimentMetaData,
                                                               title='Experiment Metadata.')

        self.TABLES['SESSION_METADETA'] = create_table_func(self.emrtFile.root.data_collection, 'session_meta_data',
                                                            SessionMetaData, title='Session Metadata.')
        self.flush()

        create_group_func(self.emrtFile.root.data_collection.events, 'experiment', title='Experiment Device Events.')
        create_group_func(self.emrtFile.root.data_collection.events, 'keyboard', title='Keyboard Device Events.')
        create_group_func(self.emrtFile.root.data_collection.events, 'mouse', title='Mouse Device Events.')
        create_group_func(self.emrtFile.root.data_collection.events, 'wintab', title='Wintab Device Events.')
        create_group_func(self.emrtFile.root.data_collection.events, 'eyetracker', title='EyeTracker Device Events.')
        create_group_func(self.emrtFile.root.data_collection.events, 'serial', title='Serial Interface Events.')
        create_group_func(self.emrtFile.root.data_collection.events, 'pstbox', title='Serial Pstbox Device Events.')

        self.flush()

    @staticmethod
    def eventTableLabel2ClassName(event_table_label):
        tokens = str(event_table_label[0] + event_table_label[1:].lower() + 'Event').split('_')
        return ''.join([t[0].upper() + t[1:] for t in tokens])

    def groupNodeForEvent(self, event_cls):
        evt_group_label = event_cls.PARENT_DEVICE.DEVICE_TYPE_STRING.lower()
        datevts_node = self.emrtFile.root.data_collection.events
        try:
            # If group for event table already exists return it....
            return datevts_node._f_get_child(evt_group_label)
        except tables.NoSuchNodeError:
            # Create the group node for the event....
            egtitle = "%s%s Device Events." % (evt_group_label[0].upper(), evt_group_label[1:])
            self.emrtFile.createGroup(datevts_node, evt_group_label, title=egtitle)
            return datevts_node._f_get_child(evt_group_label)

    def updateDataStoreStructure(self, device_instance, event_class_dict):
        dfilter = tables.Filters(complevel=0, complib='zlib', shuffle=False, fletcher32=False)

        for event_cls_name, event_cls in event_class_dict.items():
            if event_cls.IOHUB_DATA_TABLE:
                table_label = event_cls.IOHUB_DATA_TABLE
                if table_label not in self.TABLES:
                    try:
                        tc_name = self.eventTableLabel2ClassName(table_label)
                        create_table_func = getattr(self.emrtFile, create_table)
                        dc_name = device_instance.__class__.__name__
                        self.TABLES[table_label] = create_table_func(self.groupNodeForEvent(event_cls),
                                                                     tc_name,
                                                                     event_cls.NUMPY_DTYPE,
                                                                     title='%s Data' % dc_name,
                                                                     filters=dfilter.copy())
                        self.flush()
                    except tables.NodeError:
                        self.TABLES[table_label] = self.groupNodeForEvent(event_cls)._f_get_child(tc_name)
                    except Exception as e:
                        print2err('---------------ERROR------------------')
                        print2err('Exception %s in iohub.datastore.updateDataStoreStructure:' % (e.__class__.__name__))
                        print2err('\tevent_cls: {0}'.format(event_cls))
                        print2err('\tevent_cls_name: {0}'.format(event_cls_name))
                        print2err('\tevent_table_label: {0}'.format(table_label))
                        print2err('\teventTable2ClassName: {0}'.format(tc_name))
                        print2err('\tgroupNodeForEvent(event_cls): {0}'.format(self.groupNodeForEvent(event_cls)))
                        print2err('\nException:')
                        printExceptionDetailsToStdErr()
                        print2err('--------------------------------------')

                if table_label in self.TABLES:
                    self.addClassMapping(event_cls, self.TABLES[table_label])
                else:
                    print2err('---- IOHUB.DATASTORE CANNOT ADD CLASS MAPPING ----')
                    print2err('\t** TABLES missing key: {0}'.format(table_label))
                    print2err('\tevent_cls: {0}'.format(event_cls))
                    print2err('\tevent_cls_name: {0}'.format(event_cls_name))
                    print2err('\teventTableLabel2ClassName: {0}'.format(self.eventTableLabel2ClassName(table_label)))
                    print2err('----------------------------------------------')

    def addClassMapping(self, ioClass, ctable):
        cmtable = self.TABLES['CLASS_TABLE_MAPPINGS']
        names = [x['class_id'] for x in cmtable.where('(class_id == %d)' % ioClass.EVENT_TYPE_ID)]
        if len(names) == 0:
            trow = cmtable.row
            trow['class_id'] = ioClass.EVENT_TYPE_ID
            trow['class_type_id'] = 1  # Device or Event etc.
            trow['class_name'] = ioClass.__name__
            trow['table_path'] = ctable._v_pathname
            trow.append()
            self.flush()

    def createOrUpdateExperimentEntry(self, experimentInfoList):
        experiment_metadata = self.TABLES['EXPERIMENT_METADETA']
        result = [row for row in experiment_metadata.iterrows() if row['code'] == experimentInfoList[1]]
        if len(result) > 0:
            result = result[0]
            self.active_experiment_id = result['experiment_id']
            return self.active_experiment_id
        max_id = 0
        id_col = experiment_metadata.col('experiment_id')
        if len(id_col) > 0:
            max_id = np.amax(id_col)
        self.active_experiment_id = max_id + 1
        experimentInfoList[0] = self.active_experiment_id
        experiment_metadata.append([tuple(experimentInfoList), ])
        self.flush()
        return self.active_experiment_id

    def createExperimentSessionEntry(self, sessionInfoDict):
        session_metadata = self.TABLES['SESSION_METADETA']
        max_id = 0
        id_col = session_metadata.col('session_id')
        if len(id_col) > 0:
            max_id = np.amax(id_col)
        self.active_session_id = int(max_id + 1)

        values = (self.active_session_id, self.active_experiment_id, sessionInfoDict['code'], sessionInfoDict['name'],
                  sessionInfoDict['comments'], sessionInfoDict['user_variables'])

        session_metadata.append([values, ])
        self.flush()
        return self.active_session_id

    def initConditionVariableTable(
            self, experiment_id, session_id, np_dtype):
        expcv_table = None
        exp_session = [('EXPERIMENT_ID', 'i4'), ('SESSION_ID', 'i4')]
        exp_session.extend(np_dtype)
        np_dtype = []
        for npctype in exp_session:
            if isinstance(npctype[0], str):
                nv = [str(npctype[0]), ]
                nv.extend(npctype[1:])
                np_dtype.append(tuple(nv))
            else:
                np_dtype.append(npctype)

        np_dtype2 = []
        for adtype in np_dtype:
            adtype2 = []
            for a in adtype:
                if isinstance(a, bytes):
                    a = str(a, 'utf-8')
                adtype2.append(a)
            np_dtype2.append(tuple(adtype2))
        np_dtype = np_dtype2
        self._EXP_COND_DTYPE = np.dtype(np_dtype)
        try:
            expCondTableName = "EXP_CV_%d" % (experiment_id)
            experimentConditionVariableTable = getattr(self.emrtFile.root.data_collection.condition_variables,
                                                       _f_get_child)(expCondTableName)
            self.TABLES['EXP_CV'] = experimentConditionVariableTable
        except NoSuchNodeError:
            try:
                experimentConditionVariableTable = getattr(self.emrtFile, create_table)(
                    self.emrtFile.root.data_collection.condition_variables, expCondTableName, self._EXP_COND_DTYPE,
                    title='Condition Variable Values for Experiment ID %d' % experiment_id)
                self.TABLES['EXP_CV'] = experimentConditionVariableTable
                self.emrtFile.flush()
            except Exception:
                printExceptionDetailsToStdErr()
                return False
        except Exception:
            print2err('Error getting expcv_table for experiment %d, table name: %s' % (experiment_id, expCondTableName))
            printExceptionDetailsToStdErr()
            return False
        self._activeRunTimeConditionVariableTable = expcv_table
        return True

    def extendConditionVariableTable(self, experiment_id, session_id, data):
        if self._EXP_COND_DTYPE is None:
            return False
        if self.emrtFile and 'EXP_CV' in self.TABLES:
            temp = [experiment_id, session_id]
            temp.extend(data)
            data = temp
            try:
                etable = self.TABLES['EXP_CV']
                for i, d in enumerate(data):
                    if isinstance(d, (list, tuple)):
                        data[i] = tuple(d)
                np_array = np.array([tuple(data), ], dtype=self._EXP_COND_DTYPE)
                etable.append(np_array)
                self.bufferedFlush()
                return True
            except Exception:
                printExceptionDetailsToStdErr()
        return False

    def checkForExperimentAndSessionIDs(self, event=None):
        if self.active_experiment_id is None or self.active_session_id is None:
            exp_id = self.active_experiment_id
            if exp_id is None:
                exp_id = 0
            sess_id = self.active_session_id
            if sess_id is None:
                sess_id = 0
            return False
        return True

    def checkIfSessionCodeExists(self, sessionCode):
        if self.emrtFile:
            wclause = 'experiment_id == %d' % (self.active_experiment_id,)
            sessionsForExperiment = self.emrtFile.root.data_collection.session_meta_data.where(wclause)
            sessionCodeMatch = [sess for sess in sessionsForExperiment if sess['code'] == sessionCode]
            if len(sessionCodeMatch) > 0:
                return True
            return False

    def _handleEvent(self, event):
        try:
            if self.checkForExperimentAndSessionIDs(event) is False:
                return False
            etype = event[DeviceEvent.EVENT_TYPE_ID_INDEX]
            eventClass = EventConstants.getClass(etype)
            etable = self.TABLES[eventClass.IOHUB_DATA_TABLE]
            event[DeviceEvent.EVENT_EXPERIMENT_ID_INDEX] = self.active_experiment_id
            event[DeviceEvent.EVENT_SESSION_ID_INDEX] = self.active_session_id

            np_array = np.array([tuple(event), ], dtype=eventClass.NUMPY_DTYPE)
            etable.append(np_array)
            self.bufferedFlush()
        except Exception:
            print2err("Error saving event: ", event)
            printExceptionDetailsToStdErr()

    def _handleEvents(self, events):
        try:
            if self.checkForExperimentAndSessionIDs(len(events)) is False:
                return False

            event = events[0]

            etype = event[DeviceEvent.EVENT_TYPE_ID_INDEX]
            eventClass = EventConstants.getClass(etype)
            etable = self.TABLES[eventClass.IOHUB_DATA_TABLE]

            np_events = []
            for event in events:
                event[DeviceEvent.EVENT_EXPERIMENT_ID_INDEX] = self.active_experiment_id
                event[DeviceEvent.EVENT_SESSION_ID_INDEX] = self.active_session_id
                np_events.append(tuple(event))

            np_array = np.array(np_events, dtype=eventClass.NUMPY_DTYPE)
            etable.append(np_array)
            self.bufferedFlush(len(np_events))
        except ioHubError as e:
            print2err(e)
        except Exception:
            printExceptionDetailsToStdErr()

    def bufferedFlush(self, eventCount=1):
        """
        If flushCounter threshold is >=0 then do some checks. If it is < 0,
        then flush only occurs when command is sent to ioHub,
        so do nothing here.
        """
        if self.flushCounter >= 0:
            if self.flushCounter == 0:
                self.flush()
                return True
            if self.flushCounter <= self._eventCounter:
                self.flush()
                self._eventCounter = 0
                return True
            self._eventCounter += eventCount
            return False

    def flush(self):
        try:
            if self.emrtFile:
                self.emrtFile.flush()
        except tables.ClosedFileError:
            pass
        except Exception:
            printExceptionDetailsToStdErr()

    def close(self):
        self.flush()
        self._activeRunTimeConditionVariableTable = None
        self.emrtFile.close()

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass


## -------------------- Utility Functions ------------------------ ##


def close_open_data_files(verbose):
    open_files = tables.file._open_files
    clall = hasattr(open_files, 'close_all')
    if clall:
        open_files.close_all()
    else:
        are_open_files = len(open_files) > 0
        if verbose and are_open_files:
            print2err('Closing remaining open data files:')
        for fileh in open_files:
            if verbose:
                print2err('%s...' % (open_files[fileh].filename,))
            open_files[fileh].close()
            if verbose:
                print2err('done')


registered_close_open_data_files = True
atexit.register(close_open_data_files, False)


## ---------------------- Pytable Definitions ------------------- ##


class ClassTableMappings(tables.IsDescription):
    class_id = UInt32Col(pos=1)
    class_type_id = UInt32Col(pos=2)  # Device or Event etc.
    class_name = StringCol(32, pos=3)
    table_path = StringCol(128, pos=4)


class ExperimentMetaData(tables.IsDescription):
    experiment_id = UInt32Col(pos=1)
    code = StringCol(256, pos=2)
    title = StringCol(256, pos=3)
    description = StringCol(4096, pos=4)
    version = StringCol(32, pos=5)


class SessionMetaData(tables.IsDescription):
    session_id = UInt32Col(pos=1)
    experiment_id = UInt32Col(pos=2)
    code = StringCol(256, pos=3)
    name = StringCol(256, pos=4)
    comments = StringCol(4096, pos=5)
    user_variables = StringCol(16384, pos=6)  # Holds json encoded version of user variable dict for session



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Stimulus object for drawing arbitrary bitmaps that can repeat (cycle)
in either dimension. One of the main stimuli for PsychoPy.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet

from psychopy.colors import Color

pyglet.options['debug_gl'] = False
import ctypes
GL = pyglet.gl

import psychopy  # so we can get the __path__
from psychopy import logging

from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import attributeSetter
from psychopy.visual.basevisual import (BaseVisualStim, ColorMixin,
                                        ContainerMixin, TextureMixin)

import numpy


class GratingStim(BaseVisualStim, TextureMixin, ColorMixin, ContainerMixin):
    """Stimulus object for drawing arbitrary bitmaps that can repeat (cycle)
    in either dimension.

    One of the main stimuli for PsychoPy.

    Formally GratingStim is just a texture behind an optional
    transparency mask (an 'alpha mask'). Both the texture and mask can be
    arbitrary bitmaps and their combination allows an enormous variety of
    stimuli to be drawn in realtime.

    **Examples**::

        myGrat = GratingStim(tex='sin', mask='circle')  # circular grating
        myGabor = GratingStim(tex='sin', mask='gauss')  # gives a 'Gabor'

    A GratingStim can be rotated scaled and shifted in position,
    its texture can be drifted in X and/or Y and it can have a spatial
    frequency in X and/or Y (for an image file that simply draws multiple
    copies in the patch).

    Also since transparency can be controlled two GratingStims can
    combine e.g. to form a plaid.

    **Using GratingStim with images from disk (jpg, tif, png, ...)**

    Ideally texture images to be rendered should be square with
    'power-of-2' dimensions e.g. 16 x 16, 128 x 128. Any image that is
    not will be upscaled (with linear interpolation) to the nearest such
    texture by PsychoPy. The size of the stimulus should be specified in
    the normal way using the appropriate units (deg, pix, cm, ...). Be
    sure to get the aspect ratio the same as the image (if you don't want it
    stretched!).
    """

    def __init__(self,
                 win,
                 tex="sin",
                 mask="none",
                 units=None,
                 anchor="center",
                 pos=(0.0, 0.0),
                 size=None,
                 sf=None,
                 ori=0.0,
                 phase=(0.0, 0.0),
                 texRes=128,
                 rgb=None,
                 dkl=None,
                 lms=None,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 contrast=1.0,
                 opacity=None,
                 depth=0,
                 rgbPedestal=(0.0, 0.0, 0.0),
                 interpolate=False,
                 blendmode='avg',
                 name=None,
                 autoLog=None,
                 autoDraw=False,
                 maskParams=None):
        """ """  # Empty docstring. All doc is in attributes
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        for unecess in ['self', 'rgb', 'dkl', 'lms']:
            self._initParams.remove(unecess)
        # initialise parent class
        super(GratingStim, self).__init__(win, units=units, name=name,
                                          autoLog=False)

        # UGLY HACK: Some parameters depend on each other for processing.
        # They are set "superficially" here.
        # TO DO: postpone calls to _createTexture, setColor and
        # _calcCyclesPerStim whin initiating stimulus
        self.__dict__['contrast'] = 1
        self.__dict__['size'] = 1
        self.__dict__['sf'] = 1
        self.__dict__['tex'] = tex
        self.__dict__['maskParams'] = maskParams

        # initialise textures and masks for stimulus
        self._texID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._texID))
        self._maskID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._maskID))
        self.__dict__['texRes'] = texRes  # must be power of 2
        self.interpolate = interpolate

        # NB Pedestal isn't currently being used during rendering - this is a
        # place-holder
        self.rgbPedestal = val2array(rgbPedestal, False, length=3)
        # No need to invoke decorator for color updating. It is done just
        # below.
        self.colorSpace = colorSpace
        self.color = color
        if rgb != None:
            logging.warning("Use of rgb arguments to stimuli are deprecated."
                            " Please use color and colorSpace args instead")
            self.color = Color(rgb, 'rgb')
        elif dkl != None:
            logging.warning("Use of dkl arguments to stimuli are deprecated."
                            " Please use color and colorSpace args instead")
            self.color = Color(dkl, 'dkl')
        elif lms != None:
            logging.warning("Use of lms arguments to stimuli are deprecated."
                            " Please use color and colorSpace args instead")
            self.color = Color(lms, 'lms')

        # set other parameters
        self.ori = float(ori)
        self.phase = val2array(phase, False)
        self._origSize = None  # updated if an image texture is loaded
        self._requestedSize = size
        self.size = size
        self.sf = val2array(sf)
        self.pos = val2array(pos, False, False)
        self.depth = depth
        self.anchor = anchor

        self.tex = tex
        self.mask = mask
        self.contrast = float(contrast)
        self.opacity = opacity
        self.autoLog = autoLog
        self.autoDraw = autoDraw
        self.blendmode=blendmode

        # fix scaling to window coords
        self._calcCyclesPerStim()

        # generate a displaylist ID
        self._listID = GL.glGenLists(1)

        # JRG: doing self._updateList() here means MRO issues for RadialStim,
        # which inherits from GratingStim but has its own _updateList code.
        # So don't want to do the update here (= ALSO the init of RadialStim).
        # Could potentially define a BaseGrating class without
        # updateListShaders code, and have GratingStim and RadialStim
        # inherit from it and add their own _updateList stuff.
        # Seems unnecessary. Instead, simply defer the update to the
        # first .draw(), should be fast:
        # self._updateList()  # ie refresh display list
        self._needUpdate = True

        # set autoLog now that params have been initialised
        wantLog = autoLog is None and self.win.autoLog
        self.__dict__['autoLog'] = autoLog or wantLog
        if self.autoLog:
            logging.exp("Created {} = {}".format(self.name, self))

    @attributeSetter
    def sf(self, value):
        """Spatial frequency of the grating texture

        Should be a :ref:`x,y-pair <attrib-xy>` or :ref:`scalar <attrib-scalar>` or None.
        If `units` == 'deg' or 'cm' units are in cycles per deg or cm as appropriate.
        If `units` == 'norm' then sf units are in cycles per stimulus (and so SF scales with stimulus size).
        If texture is an image loaded from a file then sf=None defaults to 1/stimSize to give one cycle of the image.
        """

        # Recode phase to numpy array
        if value is None:
            # Set the sf to default (e.g. to the 1.0/size of the loaded image
            if (self.units in ('pix', 'pixels') or
                    self._origSize is not None and
                    self.units in ('deg', 'cm')):
                value = 1.0 / self.size  # default to one cycle
            else:
                value = numpy.array([1.0, 1.0])
        else:
            value = val2array(value)

        # Set value and update stuff
        self.__dict__['sf'] = value
        self._calcCyclesPerStim()
        self._needUpdate = True

    @attributeSetter
    def phase(self, value):
        """Phase of the stimulus in each dimension of the texture.

        Should be an :ref:`x,y-pair <attrib-xy>` or
        :ref:`scalar <attrib-scalar>`

        **NB** phase has modulus 1 (rather than 360 or 2*pi)
        This is a little unconventional but has the nice effect
        that setting phase=t*n drifts a stimulus at n Hz
        """
        # Recode phase to numpy array
        value = val2array(value)
        self.__dict__['phase'] = value
        self._needUpdate = True

    # overload ColorMixin methods so that they refresh the image after being called
    @property
    def foreColor(self):
        # Call setter of parent mixin
        return ColorMixin.foreColor.fget(self)
    @foreColor.setter
    def foreColor(self, value):
        # Call setter of parent mixin
        ColorMixin.foreColor.fset(self, value)
        # Reset texture
        self._needTextureUpdate = True
        self._needUpdate = True

    @property
    def contrast(self):
        # Call setter of parent mixin
        return ColorMixin.contrast.fget(self)
    @contrast.setter
    def contrast(self, value):
        # Call setter of parent mixin
        ColorMixin.contrast.fset(self, value)
        # Reset texture
        self._needTextureUpdate = True
        self._needUpdate = True

    @property
    def opacity(self):
        # Call setter of parent mixin
        return BaseVisualStim.opacity.fget(self)
    @opacity.setter
    def opacity(self, value):
        # Call setter of parent mixin
        BaseVisualStim.opacity.fset(self, value)
        # Reset texture
        self._needTextureUpdate = True
        self._needUpdate = True

    @attributeSetter
    def tex(self, value):
        """Texture to used on the stimulus as a grating (aka carrier)

        This can be one of various options:
            + **'sin'**,'sqr', 'saw', 'tri', None (resets to default)
            + the name of an image file (most formats supported)
            + a numpy array (1xN or NxN) ranging -1:1

        If specifying your own texture using an image or numpy array
        you should ensure that the image has square power-of-two dimesnions
        (e.g. 256 x 256). If not then PsychoPy will upsample your stimulus
        to the next larger power of two.
        """
        self._createTexture(value, id=self._texID,
                            pixFormat=GL.GL_RGB, stim=self,
                            res=self.texRes, maskParams=self.maskParams)
        # if user requested size=None then update the size for new stim here
        if hasattr(self, '_requestedSize') and self._requestedSize is None:
            self.size = None  # Reset size do default
        self.__dict__['tex'] = value
        self._needTextureUpdate = False

    @attributeSetter
    def blendmode(self, value):
        """The OpenGL mode in which the stimulus is draw

        Can the 'avg' or 'add'. Average (avg) places the new stimulus over the old one
        with a transparency given by its opacity. Opaque stimuli will hide other stimuli
        transparent stimuli won't. Add performs the arithmetic sum of the new stimulus and the ones
        already present.

        """
        self.__dict__['blendmode'] = value
        self._needUpdate = True

    def setSF(self, value, operation='', log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('sf', value, operation, log=log)

    def setPhase(self, value, operation='', log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('phase', value, operation, log=log)

    def setTex(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self.tex = value

    def setBlendmode(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('blendmode', value, log=log)

    def draw(self, win=None):
        """Draw the stimulus in its relevant window. You must call
        this method after every MyWin.flip() if you want the
        stimulus to appear on that frame and then update the screen
        again.
        """

        if win is None:
            win = self.win
        saveBlendMode = win.blendMode
        win.setBlendMode(self.blendmode, log=False)
        self._selectWindow(win)

        # do scaling
        GL.glPushMatrix()  # push before the list, pop after
        win.setScale('pix')
        # the list just does the texture mapping
        GL.glColor4f(*self._foreColor.render('rgba1'))

        if self._needTextureUpdate:
            self.setTex(value=self.tex, log=False)
        if self._needUpdate:
            self._updateList()
        GL.glCallList(self._listID)

        # return the view to previous state
        GL.glPopMatrix()
        win.setBlendMode(saveBlendMode, log=False)

    def _updateListShaders(self):
        """The user shouldn't need this method since it gets called
        after every call to .set() Basically it updates the OpenGL
        representation of your stimulus if some parameter of the
        stimulus changes. Call it if you change a property manually
        rather than using the .set() command
        """
        self._needUpdate = False
        GL.glNewList(self._listID, GL.GL_COMPILE)
        # setup the shaderprogram
        _prog = self.win._progSignedTexMask
        GL.glUseProgram(_prog)
        # set the texture to be texture unit 0
        GL.glUniform1i(GL.glGetUniformLocation(_prog, b"texture"), 0)
        # mask is texture unit 1
        GL.glUniform1i(GL.glGetUniformLocation(_prog, b"mask"), 1)
        # mask
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._maskID)
        GL.glEnable(GL.GL_TEXTURE_2D)  # implicitly disables 1D

        # main texture
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texID)
        GL.glEnable(GL.GL_TEXTURE_2D)

        Ltex = (-self._cycles[0] / 2) - self.phase[0] + 0.5
        Rtex = (+self._cycles[0] / 2) - self.phase[0] + 0.5
        Ttex = (+self._cycles[1] / 2) - self.phase[1] + 0.5
        Btex = (-self._cycles[1] / 2) - self.phase[1] + 0.5
        Lmask = Bmask = 0.0
        Tmask = Rmask = 1.0  # mask

        # access just once because it's slower than basic property
        vertsPix = self.verticesPix
        GL.glBegin(GL.GL_QUADS)  # draw a 4 sided polygon
        # right bottom
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Rtex, Btex)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Rmask, Bmask)
        GL.glVertex2f(vertsPix[0, 0], vertsPix[0, 1])
        # left bottom
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Ltex, Btex)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Lmask, Bmask)
        GL.glVertex2f(vertsPix[1, 0], vertsPix[1, 1])
        # left top
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Ltex, Ttex)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Lmask, Tmask)
        GL.glVertex2f(vertsPix[2, 0], vertsPix[2, 1])
        # right top
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Rtex, Ttex)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Rmask, Tmask)
        GL.glVertex2f(vertsPix[3, 0], vertsPix[3, 1])
        GL.glEnd()

        # unbind the textures
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)  # implicitly disables 1D
        # main texture
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)

        GL.glUseProgram(0)

        GL.glEndList()

    def __del__(self):
        try:
            GL.glDeleteLists(self._listID, 1)
        except Exception:
            pass  # probably we don't have a _listID property
        try:
            # remove textures from graphics card to prevent crash
            self.clearTextures()
        except Exception:
            pass

    def _calcCyclesPerStim(self):
        if self.units in ('norm', 'height'):
            # this is the only form of sf that is not size dependent
            self._cycles = self.sf
        else:
            self._cycles = self.sf * self.size
